import argparse
from spbnet.visualize.attn import attn
from pathlib import Path


class CLICommand:
    """
    Get attention score (html) according to a given weight.
    """

    @staticmethod
    def add_arguments(parser: argparse.ArgumentParser):
        parser.add_argument(
            "--cif-path",
            type=str,
            help="cif-path to get attention score",
            required=True,
        )
        parser.add_argument(
            "--modal-dir",
            type=str,
            help="directory contains modal data: graphdata, griddata8, grid",
            required=True,
        )
        parser.add_argument(
            "--ckpt",
            type=str,
            help="checkpoint(weights)",
            required=True,
        )
        parser.add_argument(
            "--out-dir",
            type=str,
            help="output directory, default is current directory",
            default=".",
        )

    @staticmethod
    def run(args):
        attn(
            Path(args.cif_path),
            Path(args.modal_dir),
            Path(args.ckpt),
            Path(args.out_dir),
        )



# MOFTransformer version 2.0.0
import random

import torch
import torch.nn as nn


class ConvLayer(nn.Module):
    """
    Convolutional operation on graphs
    (https://github.com/txie-93/cgcnn)
    """

    def __init__(self, atom_fea_len, nbr_fea_len):
        super().__init__()
        self.atom_fea_len = atom_fea_len
        self.nbr_fea_len = nbr_fea_len
        self.fc_full = nn.Linear(
            2 * self.atom_fea_len + self.nbr_fea_len, 2 * self.atom_fea_len
        )
        self.sigmoid = nn.Sigmoid()
        self.softplus1 = nn.Softplus()
        self.bn1 = nn.BatchNorm1d(2 * self.atom_fea_len)
        self.bn2 = nn.BatchNorm1d(self.atom_fea_len)
        self.softplus2 = nn.Softplus()

    def forward(self, atom_in_fea, nbr_fea, nbr_fea_idx):
        """
        Forward pass

        N: Total number of atoms in the batch
        M: Max number of neighbors

        Args:
        atom_in_fea: Variable(torch.Tensor) shape (N, atom_fea_len)
          Atom hidden features before convolution
        nbr_fea: Variable(torch.Tensor) shape (N, M, nbr_fea_len)
          Bond features of each atom's M neighbors
        nbr_fea_idx: torch.LongTensor shape (N, M)
          Indices of M neighbors of each atom

        Returns:

        atom_out_fea: nn.Variable shape (N, atom_fea_len)
          Atom hidden features after convolution

        """

        N, M = nbr_fea_idx.shape
        # convolution
        atom_nbr_fea = atom_in_fea[nbr_fea_idx, :]  # [N, M, atom_fea_len]

        total_nbr_fea = torch.cat(
            [
                atom_in_fea.unsqueeze(1).expand(N, M, self.atom_fea_len),
                # [N, atom_fea_len] -> [N, M, atom_fea_len] -> v_i
                atom_nbr_fea,  # [N, M, atom_fea_len] -> v_j
                nbr_fea,
            ],  # [N, M, nbr_fea_len] -> u(i,j)_k
            dim=2,
        )
        # [N, M, atom_fea_len*2+nrb_fea_len]

        total_gated_fea = self.fc_full(total_nbr_fea)  # [N, M, atom_fea_len*2]
        total_gated_fea = self.bn1(
            total_gated_fea.view(-1, self.atom_fea_len * 2)
        ).view(
            N, M, self.atom_fea_len * 2
        )  # [N, M, atom_fea_len*2]
        nbr_filter, nbr_core = total_gated_fea.chunk(2, dim=2)  # [N, M, atom_fea_len]
        nbr_filter = self.sigmoid(nbr_filter)
        nbr_core = self.softplus1(nbr_core)
        nbr_sumed = torch.sum(nbr_filter * nbr_core, dim=1)  # [N, atom_fea_len]
        nbr_sumed = self.bn2(nbr_sumed)
        out = self.softplus2(atom_in_fea + nbr_sumed)  # [N, atom_fea_len]
        return out


class CrystalGraphConvNet(nn.Module):
    """
    Create a crystal graph convolutional neural network for predicting total
    material properties.
    """

    def __init__(
        self,
        orig_atom_fea_len,
        nbr_fea_len,
        atom_fea_len=64,
        n_conv=3,
        h_fea_len=128,
        max_graph_len=512,
        n_h=1,
        drop_ratio=0,
    ):
        """
        Initialize CrystalGraphConvNet.
        Parameters
        ----------
        orig_atom_fea_len: int
          Number of atom features in the input.
        nbr_fea_len: int
          Number of bond features.
        atom_fea_len: int
          Number of hidden atom features in the convolutional layers
        n_conv: int
          Number of convolutional layers
        h_fea_len: int
          Number of hidden features after pooling
        n_h: int
          Number of hidden layers after pooling
        """
        super(CrystalGraphConvNet, self).__init__()

        self.max_graph_len = max_graph_len
        self.pad_embedding = nn.Embedding(1, h_fea_len)

        self.drop_ratio = drop_ratio
        self.embedding = nn.Linear(orig_atom_fea_len, atom_fea_len)
        self.convs = nn.ModuleList(
            [
                ConvLayer(atom_fea_len=atom_fea_len, nbr_fea_len=nbr_fea_len)
                for _ in range(n_conv)
            ]
        )

        self.fc_head = nn.Linear(atom_fea_len, h_fea_len)

    def forward(self, atom_fea, nbr_fea, nbr_fea_idx, crystal_atom_idx):
        """
        Forward pass
        N: Total number of atoms in the batch
        M: Max number of neighbors
        N0: Total number of crystals in the batch
        Parameters
        ----------
        atom_fea: Variable(torch.Tensor) shape (N, orig_atom_fea_len)
          Atom features from atom type
        nbr_fea: Variable(torch.Tensor) shape (N, M, nbr_fea_len)
          Bond features of each atom's M neighbors
        nbr_fea_idx: torch.LongTensor shape (N, M)
          Indices of M neighbors of each atom
        crystal_atom_idx: list of torch.LongTensor of length N0
          Mapping from the crystal idx to atom idx
        Returns
        -------
        prediction: nn.Variable shape (N, )
          Atom hidden features after convolution
        """
        atom_fea = self.embedding(atom_fea)
        for conv_func in self.convs:
            atom_fea = conv_func(atom_fea, nbr_fea, nbr_fea_idx)
        # crys_fea = self.pooling(atom_fea, crystal_atom_idx)
        # crys_fea = self.conv_to_fc(self.conv_to_fc_softplus(crys_fea))
        # crys_fea = self.conv_to_fc_softplus(crys_fea)

        return self.fc_head(atom_fea)

    def reconstruct_batch(self, atom_fea: torch.Tensor, crystal_atom_idx: torch.Tensor):
        """
        Params:
            atom_fea: [N0, h_dim]
            crystal_atom_idx: [N0]
        Return:
            atom_fea: [B, max_atom_len, h_dim]
            atom_fea_mask: [B, max_atom_len]
        """
        crystal_num = len(crystal_atom_idx)
        new_atom_fea = [[] for _ in range(crystal_num)]
        new_atom_fea_pad_mask = [None for _ in range(crystal_num)]
        for crystal_idx, atom_idxs in enumerate(crystal_atom_idx):
            for atom_idx in atom_idxs:
                new_atom_fea[crystal_idx].append(atom_fea[atom_idx])
        pad_embed = self.pad_embedding(
            torch.zeros([1], dtype=torch.long, device=atom_fea.device)
        ).view(
            -1
        )  # [h_fea_len]
        for crystal_idx, crystal_atom_fea in enumerate(new_atom_fea):
            if len(crystal_atom_fea) > self.max_graph_len:
                crystal_atom_fea = crystal_atom_fea[: self.max_graph_len]
                new_atom_fea[crystal_idx] = torch.stack(crystal_atom_fea).to(atom_fea)
                new_atom_fea_pad_mask[crystal_idx] = torch.zeros(
                    [self.max_graph_len]
                ).to(atom_fea)
            else:
                mask = torch.cat(
                    [
                        torch.zeros(len(crystal_atom_fea)),
                        torch.ones(self.max_graph_len - len(crystal_atom_fea)),
                    ]
                ).to(atom_fea)
                new_atom_fea_pad_mask[crystal_idx] = mask
                while len(crystal_atom_fea) < self.max_graph_len:
                    crystal_atom_fea.append(pad_embed.clone())
                new_atom_fea[crystal_idx] = torch.stack(crystal_atom_fea).to(atom_fea)
        new_atom_fea = torch.stack(new_atom_fea).to(atom_fea)
        new_atom_fea_pad_mask = torch.stack(new_atom_fea_pad_mask).to(atom_fea)
        return new_atom_fea, new_atom_fea_pad_mask


class GraphEmbeddings(nn.Module):
    """
    Generate Embedding layers made by only convolution layers of CGCNN (not pooling)
    (https://github.com/txie-93/cgcnn)
    """

    def __init__(
        self, atom_fea_len, nbr_fea_len, max_graph_len, hid_dim, n_conv=3, vis=False
    ):
        super().__init__()
        self.atom_fea_len = atom_fea_len
        self.nbr_fea_len = nbr_fea_len
        self.max_graph_len = max_graph_len
        self.hid_dim = hid_dim
        self.embedding = nn.Embedding(119, atom_fea_len)  # 119 -> max(atomic number)
        self.convs = nn.ModuleList(
            [
                ConvLayer(atom_fea_len=atom_fea_len, nbr_fea_len=nbr_fea_len)
                for _ in range(n_conv)
            ]
        )
        self.fc = nn.Linear(atom_fea_len, hid_dim)

        self.vis = vis

    def forward(
        self,
        atom_num,
        nbr_idx,
        nbr_fea,
        crystal_atom_idx,
        uni_idx,
        uni_count,
        moc=None,
    ):
        """
        Args:
            atom_num (tensor): [N', atom_fea_len]
            nbr_idx (tensor): [N', M]
            nbr_fea (tensor): [N', M, nbr_fea_len]
            crystal_atom_idx (list): [B]
            uni_idx (list) : [B]
            uni_count (list) : [B]
        Returns:
            new_atom_fea (tensor): [B, max_graph_len, hid_dim]
            mask (tensor): [B, max_graph_len]
        """
        assert self.nbr_fea_len == nbr_fea.shape[-1]

        atom_fea = self.embedding(atom_num)  # [N', atom_fea_len]
        for conv in self.convs:
            atom_fea = conv(atom_fea, nbr_fea, nbr_idx)  # [N', atom_fea_len]
        atom_fea = self.fc(atom_fea)  # [N', hid_dim]

        new_atom_fea, new_atom_fea_pad_mask = self.reconstruct_batch(
            atom_fea, crystal_atom_idx
        )
        # [B, max_graph_len, hid_dim], [B, max_graph_len]
        return (
            new_atom_fea,
            new_atom_fea_pad_mask,
        )  # None will be replaced with MOC

    def reconstruct_batch(self, atom_fea, crystal_atom_idx):
        # return new_atom_fea, mask, mo_label
        batch_size = len(crystal_atom_idx)

        new_atom_fea = torch.full(
            size=[batch_size, self.max_graph_len, self.hid_dim], fill_value=0.0
        ).to(atom_fea)
        new_atom_fea_pad_mask = torch.full(
            size=[batch_size, self.max_graph_len], fill_value=0, dtype=torch.int
        ).to(atom_fea)
        for crystal_idx, atom_idxs in enumerate(crystal_atom_idx):
            if len(atom_idxs) < self.max_graph_len:
                new_atom_fea_pad_mask[crystal_idx] = torch.cat(
                    [
                        torch.zeros([len(atom_idxs)], dtype=torch.int),
                        torch.ones(
                            [self.max_graph_len - len(atom_idxs)], dtype=torch.int
                        ),
                    ]
                ).to(atom_fea)
            else:
                new_atom_fea_pad_mask[crystal_idx] = torch.zeros(
                    [self.max_graph_len], dtype=torch.int
                ).to(atom_fea)
            idx_ = atom_idxs[: self.max_graph_len]
            new_atom_fea[crystal_idx][: len(idx_)] = atom_fea[idx_]

        return new_atom_fea, new_atom_fea_pad_mask



import pickle
import pandas as pd

import numpy as np

import torch
from pathlib import Path
from tqdm import tqdm
from functools import lru_cache

import json


class Dataset(torch.utils.data.Dataset):
    def __init__(
        self,
        df: pd.DataFrame,
        data_dir: Path,
        nbr_fea_len: int,
        draw_false_grid=False,
    ):
        """
        Dataset for MOF.
        Args:
            df (pd.DataFrame): cifid, mofid, target
            data_dir (str): where dataset cif files and energy grid file; exist via model.utils.prepare_data.py
            split (str) : train, test, split
            draw_false_grid (int, optional):  how many generating false_grid_data
            nbr_fea_len (int) : nbr_fea_len for gaussian expansion
        """
        super().__init__()
        self.df = df
        # calc topo
        self.topo2tid = None
        # topofile = Path('/cpfs01/projects-HDD/neikiuyiliaodamoxing_HDD/sunyuqi/zjw/SpbNet/data/topo2tid.json')
        topofile = data_dir / "topo2tid.json"
        if topofile.exists():
            with topofile.open("r") as f:
                self.topo2tid = json.load(f)
        else:
            topo2tid = dict()
            for item in self.df.iloc:
                topo2tid[item["topo"]] = True
            self.topo2tid = dict()
            for tid, topo in enumerate(topo2tid.keys()):
                self.topo2tid[topo] = tid
            with topofile.open("w") as f:
                json.dump(self.topo2tid, f)

        # self.cifids = df["cifid"]
        self.data_dir = data_dir
        self.draw_false_grid = draw_false_grid
        self.nbr_fea_len = nbr_fea_len

        self.max_atom_num = 512

        # self.items = []
        # print("Start to load dataset")
        # for i in tqdm(range(len(self.df)), total=len(self.df)):
        #     self.items.append(self.getitem(i))

    def __len__(self):
        return len(self.df)

    def get_topo_num(self):
        return len(list(self.topo2tid.keys()))

    # grid
    @lru_cache(maxsize=-1)
    def load_griddata(self, file_griddata: Path):
        return np.load(file_griddata.absolute())

    def handle_griddata(self, file_griddata: Path, GRID=30, channel=20):
        griddata = self.load_griddata(file_griddata)  # [30*30*30, 20] uint8
        griddata = griddata.astype(np.float32)
        # # original
        # griddata = griddata / 250 * 101 + 1

        griddata = griddata.reshape(GRID, GRID, GRID, channel)
        lj = griddata[:, :, :, 12]
        correction = np.concatenate(
            [griddata[:, :, :, :12], griddata[:, :, :, 13:19]], axis=-1
        )

        return torch.from_numpy(lj).float(), torch.from_numpy(correction).float()

    @staticmethod
    def calculate_volume(a, b, c, angle_a, angle_b, angle_c):
        a_ = np.cos(angle_a * np.pi / 180)
        b_ = np.cos(angle_b * np.pi / 180)
        c_ = np.cos(angle_c * np.pi / 180)

        v = a * b * c * np.sqrt(np.abs(1 - a_**2 - b_**2 - c_**2 + 2 * a_ * b_ * c_))

        return v.item() / (60 * 60 * 60)  # normalized volume

    def get_raw_grid_data(self, cifid):
        file_grid = self.data_dir / "grid" / f"{cifid}.grid"
        file_griddata = self.data_dir / "griddata8" / f"{cifid}.npy"

        # get grid
        with file_grid.open("r") as f:
            lines = f.readlines()
            a, b, c = [float(i) for i in lines[0].split()[1:]]
            angle_a, angle_b, angle_c = [float(i) for i in lines[1].split()[1:]]
            cell = [int(i) for i in lines[2].split()[1:]]

        volume = self.calculate_volume(a, b, c, angle_a, angle_b, angle_c)

        # get grid data
        lj, corr = self.handle_griddata(file_griddata)

        return cell, volume, lj, corr

    def get_grid_data(self, cifid, draw_false_grid=False):
        cell, volume, lj, corr = self.get_raw_grid_data(cifid)
        return {"cell": cell, "volume": volume, "lj": lj, "corr": corr}

    @staticmethod
    def get_gaussian_distance(distances, num_step, dmax, dmin=0, var=0.2):
        """
        Expands the distance by Gaussian basis
        (https://github.com/txie-93/cgcnn.git)
        """

        assert dmin < dmax
        _filter = np.linspace(
            dmin, dmax, num_step
        )  # = np.arange(dmin, dmax + step, step) with step = 0.2

        return np.exp(-((distances[..., np.newaxis] - _filter) ** 2) / var**2).float()

    @lru_cache(maxsize=-1)
    def load_graphdata(self, file_graph: Path):
        return pickle.load(file_graph.open("rb"))

    # @lru_cache(maxsize=None)
    def get_graph(self, cifid: str):
        # moftransformer
        file_graph = self.data_dir / "graphdata" / f"{cifid}.graphdata"

        graphdata = self.load_graphdata(file_graph)
        # graphdata = ["cifid", "atom_num", "nbr_idx", "nbr_dist", "uni_idx", "uni_count"]
        atom_num = torch.LongTensor(graphdata[1].copy())
        nbr_idx = torch.LongTensor(graphdata[2].copy()).view(len(atom_num), -1)
        nbr_dist = torch.FloatTensor(graphdata[3].copy()).view(len(atom_num), -1)

        nbr_fea = torch.FloatTensor(
            self.get_gaussian_distance(nbr_dist, num_step=self.nbr_fea_len, dmax=8)
        )

        uni_idx = graphdata[4]
        uni_count = graphdata[5]

        return {
            "atom_num": atom_num,
            "nbr_idx": nbr_idx,
            "nbr_fea": nbr_fea,
            "uni_idx": uni_idx,
            "uni_count": uni_count,
        }

    @lru_cache(maxsize=-1)
    def load_atomgrid(self, file_atomgrid: Path):
        return np.load(file_atomgrid)

    # @lru_cache(maxsize=-1)
    def get_atomgrid(self, cifid: str):
        file_atomgrid = self.data_dir / "atomgrid" / f"{cifid}.npy"
        arr = self.load_atomgrid(file_atomgrid)
        tensor = torch.from_numpy(arr).float()
        # tensor = torch.load(file_atomgrid)
        return {"atomgrid": tensor}

    def get_coulomb(self, cifid: str):
        file_coulomb = self.data_dir / "coulomb" / f"{cifid}.coulomb"
        potential = torch.load(file_coulomb, map_location="cpu")
        return {"coulomb": potential}

    def __getitem__(self, index):
        ret = dict()

        item = self.df.iloc[index]
        cifid = item["cifid"]
        topo = self.topo2tid[item["topo"]]
        voidfraction = item["voidfraction"]

        ret.update({"cifid": cifid, "topo": topo, "voidfraction": voidfraction})
        # ret.update(self.get_atom(cifid))
        ret.update(self.get_grid_data(cifid, draw_false_grid=self.draw_false_grid))
        ret.update(self.get_graph(cifid))
        ret.update(self.get_atomgrid(cifid))
        # ret.update(self.get_coulomb(cifid))

        # ret.update(self.get_tasks(index))

        return ret

    # def __getitem__(self, index):
    #     return self.items(index)
    # ret = dict()

    # item = self.df.iloc[index]
    # cifid = item["cifid"]
    # topo = self.topo2tid[item["topo"]]
    # voidfraction = item["voidfraction"]

    # ret.update({"cifid": cifid, "topo": topo, "voidfraction": voidfraction})
    # # ret.update(self.get_atom(cifid))
    # ret.update(self.get_grid_data(cifid, draw_false_grid=self.draw_false_grid))
    # ret.update(self.get_graph(cifid))
    # ret.update(self.get_atomgrid(cifid))
    # # ret.update(self.get_coulomb(cifid))

    # # ret.update(self.get_tasks(index))

    # return ret

    @staticmethod
    def collate(batch, img_size):
        """
        collate batch
        Args:
            batch (dict): [cifid, atom_num, nbr_idx, nbr_fea, uni_idx, uni_count,
                            grid_data, cell, (false_grid_data, false_cell), target]
            img_size (int): maximum length of img size

        Returns:
            dict_batch (dict): [cifid, atom_num, nbr_idx, nbr_fea, crystal_atom_idx,
                                uni_idx, uni_count, grid, false_grid_data, target]
        """
        batch_size = len(batch)
        keys = set([key for b in batch for key in b.keys()])

        dict_batch = {k: [dic[k] if k in dic else None for dic in batch] for k in keys}

        # graph
        # moftransformer
        batch_atom_num = dict_batch["atom_num"]
        batch_nbr_idx = dict_batch["nbr_idx"]
        batch_nbr_fea = dict_batch["nbr_fea"]

        crystal_atom_idx = []
        base_idx = 0
        for i, nbr_idx in enumerate(batch_nbr_idx):
            n_i = nbr_idx.shape[0]
            crystal_atom_idx.append(torch.arange(n_i) + base_idx)
            nbr_idx += base_idx
            base_idx += n_i

        dict_batch["atom_num"] = torch.cat(batch_atom_num, dim=0)
        dict_batch["nbr_idx"] = torch.cat(batch_nbr_idx, dim=0)
        dict_batch["nbr_fea"] = torch.cat(batch_nbr_fea, dim=0)
        dict_batch["crystal_atom_idx"] = crystal_atom_idx

        # atom grid
        batch_atomgrid = dict_batch["atomgrid"]
        batch_atomgrid = torch.stack(batch_atomgrid, dim=0)
        dict_batch["atomgrid"] = (
            batch_atomgrid  # [B, GRID, GRID, GRID] aka [B, 30, 30, 30]
        )

        # target
        # batch_target = dict_batch["target"]
        # batch_target = torch.tensor(batch_target, dtype=torch.float)
        # dict_batch["target"] = batch_target
        batch_topo = dict_batch["topo"]
        batch_topo = torch.tensor(batch_topo, dtype=torch.long)
        dict_batch["topo"] = batch_topo
        batch_voidfraction = dict_batch["voidfraction"]
        batch_voidfraction = torch.tensor(batch_voidfraction, dtype=torch.float)
        dict_batch["voidfraction"] = batch_voidfraction

        # # coulomb
        # batch_coulomb = dict_batch["coulomb"]  # List[[H, W, D]]
        # batch_coulomb = torch.stack(batch_coulomb)  # [B, H, W, D]
        # batch_coulomb.unsqueeze_(1)
        # # batch_coulomb = torch.clamp(batch_coulomb, min=None, max=300)
        # batch_coulomb = normalize(batch_coulomb)
        # dict_batch["coulomb"] = batch_coulomb

        # grid
        batch_lj_data = dict_batch["lj"]
        batch_lj_data = torch.stack(batch_lj_data)
        batch_lj_data = batch_lj_data.view(batch_lj_data.shape[0], 30, 30, 30)
        dict_batch["lj"] = batch_lj_data

        batch_corr_data = dict_batch["corr"]
        batch_corr_data = torch.stack(batch_corr_data)
        batch_corr_data = batch_corr_data.view(batch_corr_data.shape[0], 30, 30, 30, 18)
        dict_batch["corr"] = batch_corr_data

        return dict_batch



from colorama import Fore, Style
from pprint import pprint as pp

import os


# 获取终端宽度
def get_terminal_width():
    try:
      term_size = os.get_terminal_size().columns
    except:
      term_size = 50
    return term_size


terminal_width = get_terminal_width()


def title(sentence: str, len=terminal_width, charac="="):
    print(
        Fore.YELLOW
        + (" SpbNet: " + sentence + " ").center(len, charac)
        + Style.RESET_ALL
    )


def err(log: str):
    print(Fore.RED + 'ERROR: ' + log + Style.RESET_ALL)


def warn(log: str):
    print(Fore.YELLOW + "WARNING: " + log + Style.RESET_ALL)


def end(log: str):
    print(Fore.BLUE + log + Style.RESET_ALL)


def start(log: str):
    print(Fore.CYAN + log + Style.RESET_ALL)


def param(**kwargs):
    pp(kwargs)



import argparse
from spbnet.data.buildAtomGrid import buildAtomGrid


class CLICommand:
    """
    Build Atom Grid data for pretrain spbnet

    Please provide directory like:

    - root_dir
        - cif
            - mof1.cif
            - mof2.cif
            ...

    This command will produce:

    - root_dir
        - cif
            - mof1.cif
            - mof2.cif
            ...
        - spbnet
            - atomgrid
                - mof1.npy
                - mof2.npy
                ...
    """

    @staticmethod
    def add_arguments(parser: argparse.ArgumentParser):
        parser.add_argument(
            "--root-dir", type=str, help="root_dir to build atom grid", required=True
        )
        parser.add_argument(
            "--cif-dir",
            type=str,
            help='cif_dir to use, default is "cif", this command will use cif files from root_dir/cif_dir',
            default="cif",
        )
        parser.add_argument(
            "--target-dir",
            type=str,
            help="target directory to build data, this command will produce data in root_dir/target_dir",
            default="spbnet",
        )
        parser.add_argument(
            "--n-process",
            type=int,
            help="multiprocessing, number of processes to use, default is 1",
            default=1,
        )

    @staticmethod
    def run(args):
        buildAtomGrid(args.root_dir, args.cif_dir, args.target_dir, args.n_process)



from pathlib import Path
import pandas as pd
from tqdm import tqdm
import pickle
import numpy as np
from spbnet.utils.echo import title, start, end, param, warn


def filterData(root_dir, modal_dir, prop_dir, prop_name, outlier):
    title("FILTER DATA")

    root_dir = Path(root_dir)
    modal_dir = root_dir / modal_dir
    prop_dir = root_dir / prop_dir
    param(
        root_dir=root_dir.absolute(),
        modal_dir=modal_dir.absolute(),
        prop_dir=prop_dir.absolute(),
        prop_name=prop_name,
    )

    def check_cifid(cifid: str):
        try:
            GRID = 30
            channel = 20

            data_dir = modal_dir
            graphdata = pickle.load(
                (data_dir / "graphdata" / f"{cifid}.graphdata").open("rb")
            )
            # graphdata = pickle.load(open(f"./spbnet/graphdata/{cifid}.graphdata", "rb"))
            file_grid = data_dir / "grid" / f"{cifid}.grid"
            file_griddata = data_dir / "griddata8" / f"{cifid}.npy"

            # get grid
            with file_grid.open("r") as f:
                lines = f.readlines()
                a, b, c = [float(i) for i in lines[0].split()[1:]]
                angle_a, angle_b, angle_c = [float(i) for i in lines[1].split()[1:]]
                cell = [int(i) for i in lines[2].split()[1:]]

            griddata = np.load(file_griddata.absolute())  # [30*30*30, 20] uint8
            griddata = griddata.astype(np.float32)

            griddata = griddata.reshape(GRID, GRID, GRID, channel)

            return True
        except:
            return False

    def get_value(df: pd.DataFrame, item, task: str):
        arr = df[task]
        q1 = np.percentile(arr, 25)
        q3 = np.percentile(arr, 75)
        iqr = q3 - q1
        if outlier <= 0:
            return item[task], True

        low = q1 - outlier * iqr
        up = q3 + outlier * iqr

        if item[task] > up:
            warn(
                f"outlier point found when filtering data, cif: {item['cifid']}, task: {task}, value: {item[task]}. The Q1 (First Quartile) of this task is {q1}, Q3 (Third Quartile) is {q3}. To avoid significant errors, this data will be ignored."
            )
            # return np.percentile(df[task], 90)
            return np.nan, False
        if item[task] < low:
            warn(
                f"outlier point found when filtering data, cif: {item['cifid']}, task: {task}, value: {item[task]}. The Q1 (First Quartile) of this task is {q1}, Q3 (Third Quartile) is {q3}. To avoid significant errors, this data will be ignored for all models in this paper (including baseline model)."
            )
            # return np.percentile(df[task], 10)
            return np.nan, False
        return item[task], True

    start("Start to filter data")
    df = pd.read_csv((prop_dir / f"{prop_name}.csv").absolute())
    tasks = [str(column) for column in df.columns[1:]]
    param(tasks=tasks)
    cifids = df["cifid"]
    items = []
    outlier_count = 0
    for item in tqdm(df.iloc, total=len(df)):
        cifid = item["cifid"]
        for task in tasks:
            if get_value(df, item, task)[1] is not True:
                outlier_count += 1
                break
        targets = [get_value(df, item, task)[0] for task in tasks]
        if not check_cifid(cifid):
            continue
        items.append([cifid] + targets)
    end(f"Filter end, All: {len(df)}, Filtered: {len(items)}, Outlier: {outlier_count}")
    start("Start to split train/validate/test dataset")
    # print(['cifid']+tasks, items[0])
    items = pd.DataFrame(items, columns=["cifid"] + tasks)
    items = items.sample(n=len(items))
    if len(items) > 7000:
        items = items.sample(7000)
    items = items.sample(n=len(items))
    train_ratio = 0.8
    val_ratio = 0.1
    size = len(items)
    items.to_csv((prop_dir / f"{prop_name}.filter.csv").absolute(), index=False)
    items[: int(size * train_ratio)].to_csv(
        (prop_dir / f"{prop_name}.train.csv").absolute(), index=False
    )
    items[int(size * train_ratio) : int(size * (train_ratio + val_ratio))].to_csv(
        (prop_dir / f"{prop_name}.validate.csv").absolute(), index=False
    )
    items[int(size * (train_ratio + val_ratio)) :].to_csv(
        (prop_dir / f"{prop_name}.test.csv").absolute(), index=False
    )
    end("Split end")
    title("FILTER DATA END")



import os
import math
import logging
import logging.handlers
import json
import subprocess
import hashlib
import pickle
import shutil
from pathlib import Path
from collections import namedtuple
from collections.abc import Iterable

import numpy as np

from tqdm import tqdm
from pymatgen.io.cif import CifParser
from pymatgen.io.ase import AseAtomsAdaptor
from pymatgen.io.cssr import Cssr

from ase.io import read
from ase.neighborlist import natural_cutoffs
from ase import neighborlist
from ase.build import make_supercell


cur_dir = Path(__file__).parent.absolute()

root_dir = cur_dir.parent
data_dir = root_dir / "data"

GRIDAY_PATH = os.path.join(data_dir.absolute(), "libs/GRIDAYS/scripts/grid_gen")
FF_PATH = os.path.join(data_dir.absolute(), "libs/GRIDAYS/FF")


def get_logger(filename):
    logger = logging.getLogger(filename)
    logger.setLevel(logging.INFO)

    formatter = logging.Formatter(
        fmt="%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )

    file_handler = logging.FileHandler(filename)
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)

    return logger


def get_unique_atoms(atoms):
    # get graph
    cutoff = natural_cutoffs(atoms)
    neighbor_list = neighborlist.NeighborList(
        cutoff, self_interaction=True, bothways=True
    )
    neighbor_list.update(atoms)
    matrix = neighbor_list.get_connectivity_matrix()

    # Get N, N^2
    numbers = atoms.numbers
    number_sqr = np.multiply(numbers, numbers)

    matrix_sqr = matrix.dot(matrix)
    matrix_cub = matrix_sqr.dot(matrix)
    matrix_sqr.data[:] = 1  # count 1 for atoms

    # calculate
    list_n = [numbers, number_sqr]
    list_m = [matrix, matrix_sqr, matrix_cub]

    arr = [numbers]

    for m in list_m:
        for n in list_n:
            arr.append(m.dot(n))

    arr = np.vstack(arr).transpose()

    uni, unique_idx, unique_count = np.unique(
        arr, axis=0, return_index=True, return_counts=True
    )

    # sort
    final_uni = uni[np.argsort(-unique_count)].tolist()
    final_unique_count = unique_count[np.argsort(-unique_count)].tolist()

    arr = arr.tolist()
    final_unique_idx = []
    for u in final_uni:
        final_unique_idx.append([i for i, a in enumerate(arr) if a == u])

    return final_unique_idx, final_unique_count


def get_crystal_graph(atoms, radius=8, max_num_nbr=12):
    dist_mat = atoms.get_all_distances(mic=True)
    nbr_mat = np.where(dist_mat > 0, dist_mat, 1000)  # 1000 is mamium number
    nbr_idx = []
    nbr_dist = []
    for row in nbr_mat:
        idx = np.argsort(row)[:max_num_nbr]
        nbr_idx.extend(idx)
        nbr_dist.extend(row[idx])

    # get same-topo atoms
    uni_idx, uni_count = get_unique_atoms(atoms)

    # convert to small size
    atom_num = np.array(list(atoms.numbers), dtype=np.int8)
    nbr_idx = np.array(nbr_idx, dtype=np.int16)
    nbr_dist = np.array(nbr_dist, dtype=np.float32)
    uni_count = np.array(uni_count, dtype=np.int16)
    return atom_num, nbr_idx, nbr_dist, uni_idx, uni_count


def _calculate_scaling_matrix_for_orthogonal_supercell(cell_matrix, eps=0.01):
    """
    cell_matrix: contains lattice vector as column vectors.
                 e.g. cell_matrix[:, 0] = a.
    eps: when value < eps, the value is assumed as zero.
    """
    inv = np.linalg.inv(cell_matrix)

    # Get minimum absolute values of each row.
    abs_inv = np.abs(inv)
    mat = np.where(abs_inv < eps, np.full_like(abs_inv, 1e30), abs_inv)
    min_values = np.min(mat, axis=1)

    # Normalize each row with minimum absolute value of each row.
    normed_inv = inv / min_values[:, np.newaxis]

    # Calculate scaling_matrix.
    # New cell = np.dot(scaling_matrix, cell_matrix).
    scaling_matrix = np.around(normed_inv).astype(np.int32)

    return scaling_matrix


def make_float16_griddata(file_griddata):
    griddata = np.fromfile(file_griddata, dtype=np.float32)
    griddata[griddata > 6e4] = 6e4
    griddata[griddata < -6e4] = -6e4
    griddata = griddata.astype(np.float16)
    return griddata


def get_energy_grid(atoms, cif_id, root_dataset, eg_logger):
    # Before 1.1.1 version : num_grid = [str(round(cell)) for cell in structure.lattice.abc]
    # After 1.1.1 version : num_grid = [30, 30, 30]
    global GRIDAY_PATH, FF_PATH

    eg_file = os.path.join(root_dataset, cif_id)
    # random_str = str(np.random.rand()).encode()
    tmp_file = os.path.join(root_dataset, f"{cif_id}.cssr")

    try:
        structure = AseAtomsAdaptor().get_structure(atoms)
        Cssr(structure).write_file(tmp_file)
        if not os.path.exists(tmp_file):
            eg_logger.info(f"{cif_id} cssr write fail")
            return False
        num_grid = ["30", "30", "30"]
        print("GridGenPath: ", GRIDAY_PATH)
        proc = subprocess.Popen(
            [
                GRIDAY_PATH,
                *num_grid,
                f"{FF_PATH}/UFF_Type.def",
                f"{FF_PATH}/UFF_FF.def",
                tmp_file,
                eg_file,
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        # proc.wait()
        out, err = proc.communicate()
    finally:
        # remove temp_file
        if os.path.exists(tmp_file):
            os.remove(tmp_file)

    if err:
        eg_logger.info(f"{cif_id} energy grid failed {err}")
        return False
    else:
        eg_logger.info(f"{cif_id} energy grid success")

    return True


def _split_dataset(root_dataset: Path, **kwargs):
    """
    make train_{task}.json, test_{task}.json, and val_{task}.json from raw_{task}.json
    :param root_cifs: root for cif files
    :param root_dataset: root for generated datasets
    :param single_task: name of downstream tasks
    :param kwargs:
        - overwrite_json (bool) : If True, overwrite {split}_task.json file when it exists. (default : False)
        - seed : (int) random seed for split data. (default : 42)
        - duplicate : (bool) If True, allow duplication of data in train, test, and validation. (default: False)
        - train_fraction : (float) fraction for train dataset. train_fraction + test_fraction must be smaller than 1 (default : 0.8)
        - test_fraction : (float) fraction for test dataset. train_fraction + test_fraction must be smaller than 1 (default : 0.1)

    :return:
    """
    # get argument from kwargs
    seed = kwargs.get("seed", 42)
    threshold = kwargs.get("threshold", 0.01)
    train_fraction = kwargs.get("train_fraction", 0.8)
    test_fraction = kwargs.get("test_fraction", 0.1)

    # get directories
    total_dir = root_dataset / "total"
    assert total_dir.exists()

    split_dir = {split: root_dataset / split for split in ["train", "test", "val"]}
    for direc in split_dir.values():
        direc.mkdir(exist_ok=True)

    # get success prepare-data list
    cif_list = {}
    CifPath = namedtuple("CifPath", ["cif", "graphdata", "grid", "griddata16"])
    for cif in total_dir.glob("*.cif"):
        cif_id = cif.stem
        graphdata = cif.with_suffix(".graphdata")
        grid = cif.with_suffix(".grid")
        griddata = cif.with_suffix(".griddata16")

        if cif.exists() and graphdata.exists() and grid.exists() and griddata.exists():
            cif_list[cif_id] = CifPath(cif, graphdata, grid, griddata)

    # get number of split
    if train_fraction + test_fraction > 1:
        raise ValueError(
            f'"train_fraction + test_fraction" must be smaller than 1.0, not {train_fraction + test_fraction}'
        )

    n_total = len(cif_list.keys())
    n_train = int(n_total * train_fraction)
    n_test = int(n_total * test_fraction)
    n_val = n_total - n_train - n_test
    n_split = {"train": n_train, "test": n_test, "val": n_val}

    # remove already-divided values
    for split, direc in split_dir.items():
        split_cifs = {cif.stem for cif in direc.glob("*.cif")}
        for cif in split_cifs:
            if cif in cif_list:
                del cif_list[cif]
                n_split[split] -= 1

    assert sum(n_split.values()) == len(cif_list), "Error! contact with code writer!"
    if not cif_list:  # NO additional divided task
        return

    for split, n in n_split.items():
        if n < -n_total * threshold:
            raise ValueError(
                "{split} folder's cif number is larger than {split}_fraction. change argument {split}_fraction."
            )

    # random split index
    cif_name = sorted(list(cif_list.keys()))
    split_idx = (
        ["train"] * n_split["train"]
        + ["test"] * n_split["test"]
        + ["val"] * n_split["val"]
    )
    np.random.seed(seed=seed)
    np.random.shuffle(split_idx)

    assert len(cif_name) == len(split_idx)

    for cif, split in zip(cif_name, split_idx):
        cifpath = cif_list[cif]
        for suffix in ["cif", "graphdata", "grid", "griddata16"]:
            src = getattr(cifpath, suffix)
            dest = root_dataset / split
            shutil.copy(src, dest)


def _split_json(root_cifs: Path, root_dataset: Path, downstream: str):
    with open(str(root_cifs / f"raw_{downstream}.json")) as f:
        src = json.load(f)
        src = {
            i.replace(".cif", ""): v for i, v in src.items()
        }  # if *.cif in JSON files

    for split in ["train", "test", "val"]:
        cif_folder = root_dataset / split
        cif_list = [cif.stem for cif in cif_folder.glob("*.cif")]
        split_json = {i: src[i] for i in cif_list if i in src}
        with open(str(root_dataset / f"{split}_{downstream}.json"), "w") as f:
            json.dump(split_json, f)


def _make_supercell(atoms, cutoff):
    """
    make atoms into supercell when cell length is less than cufoff (min_length)
    """
    # when the cell lengths are smaller than radius, make supercell to be longer than the radius
    scale_abc = []
    for l in atoms.cell.cellpar()[:3]:
        if l < cutoff:
            scale_abc.append(math.ceil(cutoff / l))
        else:
            scale_abc.append(1)

    # make supercell
    m = np.zeros([3, 3])
    np.fill_diagonal(m, scale_abc)
    atoms = make_supercell(atoms, m)
    return atoms


def make_prepared_data(
    cif: Path, root_dataset_total: Path, logger=None, eg_logger=None, **kwargs
):
    if logger is None:
        logger = get_logger(filename="prepare_data.log")
    if eg_logger is None:
        eg_logger = get_logger(filename="prepare_energy_grid.log")

    if isinstance(cif, str):
        cif = Path(cif)
    if isinstance(root_dataset_total, str):
        root_dataset_total = Path(root_dataset_total)

    root_dataset_total.mkdir(exist_ok=True, parents=True)

    max_length = kwargs.get("max_length", 120.0)
    min_length = kwargs.get("min_length", 30.0)
    max_num_nbr = kwargs.get("max_num_nbr", 12)
    max_num_unique_atoms = kwargs.get("max_num_unique_atoms", 300)
    max_num_atoms = kwargs.get("max_num_atoms", None)

    cif_id: str = cif.stem

    p_graphdata = root_dataset_total / f"{cif_id}.graphdata"
    p_griddata = root_dataset_total / f"{cif_id}.griddata16"
    p_grid = root_dataset_total / f"{cif_id}.grid"

    # # Grid data and Graph data already exists
    # if p_graphdata.exists() and p_griddata.exists() and p_grid.exists():
    #     logger.info(f"{cif_id} graph data already exists")
    #     eg_logger.info(f"{cif_id} energy grid already exists")
    #     return True

    # valid cif check
    try:
        CifParser(cif).get_structures()
    except ValueError as e:
        logger.info(f"{cif_id} failed : {e} (error when reading cif with pymatgen)")
        return False

    # read cif by ASE
    try:
        atoms = read(str(cif))
    except Exception as e:
        logger.error(f"{cif_id} failed : {e}")
        return False

    # 1. get crystal graph
    atoms = _make_supercell(atoms, cutoff=8)  # radius = 8

    if max_num_atoms and len(atoms) > max_num_atoms:
        logger.error(
            f"{cif_id} failed : number of atoms are larger than `max_num_atoms` ({max_num_atoms})"
        )
        return False

    atom_num, nbr_idx, nbr_dist, uni_idx, uni_count = get_crystal_graph(
        atoms, radius=8, max_num_nbr=max_num_nbr
    )
    if len(nbr_idx) < len(atom_num) * max_num_nbr:
        logger.error(
            f"{cif_id} failed : num_nbr is smaller than max_num_nbr. please make radius larger"
        )
        return False

    if len(uni_idx) > max_num_unique_atoms:
        logger.error(
            f"{cif_id} failed : The number of topologically unique atoms is larget than `max_num_unique_atoms` ({max_num_unique_atoms})"
        )
        return False

    # 2. make supercell with min_length
    atoms_eg = _make_supercell(atoms, cutoff=min_length)
    for l in atoms_eg.cell.cellpar()[:3]:
        if l > max_length:
            logger.error(f"{cif_id} failed : supercell have more than max_length")
            return False

    # 3. calculate energy grid
    eg_success = get_energy_grid(atoms_eg, cif_id, root_dataset_total, eg_logger)

    if eg_success:
        logger.info(f"{cif_id} succeed : supercell length {atoms.cell.cellpar()[:3]}")

        # save cif files
        save_cif_path = root_dataset_total / f"{cif_id}.cif"
        atoms.write(filename=save_cif_path)

        # save graphdata file
        data = [cif_id, atom_num, nbr_idx, nbr_dist, uni_idx, uni_count]
        with open(str(p_graphdata), "wb") as f:
            pickle.dump(data, f)
        return True
    else:
        return False


def prepare_data(root_cifs, root_dataset, downstream, **kwargs):
    """
    Args:
        root_cifs (str): root for cif files,
                        it should contains "train" and "test" directory in root_cifs
                        ("val" directory is optional)
        root_dataset (str): root for generated datasets
        downstream (str or list) : name of downstream tasks

    kwargs:
        - seed : (int) random seed for split data. (default : 42)
        - train_fraction : (float) fraction for train dataset. train_fraction + test_fraction must be smaller than 1 (default : 0.8)
        - test_fraction : (float) fraction for test dataset. train_fraction + test_fraction must be smaller than 1 (default : 0.1)

        - get_primitive (bool) : If True, use primitive cell in graph embedding
        - max_num_unique_atoms (int): max number unique atoms in primitive cells (default: 300)
        - max_num_supercell_atoms (int or None): max number atoms in super cell atoms (default: None)
        - max_length (float) : maximum length of supercell
        - min_length (float) : minimum length of supercell
        - max_num_nbr (int) : maximum number of neighbors when calculating graph
    """
    if not os.path.exists(GRIDAY_PATH):
        raise ImportError(
            "GRIDAY must be installed. \n"
            "Run the following code in bash, \n\n"
            "$ moftransformer install-griday\n\n"
            "or run the following code on Python\n\n"
            ">>> from moftransformer.utils import install_griday\n"
            ">>> install_griday()"
        )

    # set logger
    logger = get_logger(filename="prepare_data.log")
    eg_logger = get_logger(filename="prepare_energy_grid.log")

    # directory to "Path"
    root_cifs = Path(root_cifs)
    root_dataset = Path(root_dataset)

    if not root_cifs.exists():
        raise ValueError(f"{root_cifs} does not exists.")

    # make prepare_data in 'total' directory
    root_dataset_total = Path(root_dataset) / "total"
    root_dataset_total.mkdir(exist_ok=True, parents=True)

    # make *.grid, *.griddata16, and *.graphdata file
    for cif in tqdm(
        root_cifs.glob("*.cif"), total=sum(1 for _ in root_cifs.glob("*.cif"))
    ):
        make_prepared_data(cif, root_dataset_total, logger, eg_logger, **kwargs)

    # automatically split data
    _split_dataset(root_dataset, **kwargs)

    # split json file
    if isinstance(downstream, str):
        _split_json(root_cifs, root_dataset, downstream)
    elif isinstance(downstream, Iterable):
        for single_downstream in downstream:
            _split_json(root_cifs, root_dataset, single_downstream)
    else:
        raise TypeError(f"task must be str or Iterable, not {type(downstream)}")



import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from sklearn.metrics import r2_score
from scipy.stats import pearsonr

plt.rcParams["font.family"] = "serif"
plt.rcParams["font.serif"] = ["Arial"] + plt.rcParams["font.serif"]
plt.rcParams["font.size"] = 12

fig, ax = plt.subplots()

datasets = [
    "pretrain",
    "coremof",
    "cof",
    "ppn",
    "zeolite",
]


def get_err(dataset: str):
    labels = np.load(f"./feat/{dataset}/agc_labels.npy").reshape(-1)
    preds = np.load(f"./feat/{dataset}/agc_preds.npy").reshape(-1)


    # idxes = np.where(labels > 0)[0]
    # labels = labels[idxes]
    # preds = preds[idxes]

    # idxes = np.where(labels < 10)[0]
    # labels = labels[idxes]
    # preds = preds[idxes]

    print(
        f"{dataset}: {r2_score(labels, preds)}, {pearsonr(labels, preds)[0]}, ratio: {np.sum(labels < 1) / len(labels)}"
    )

    rand_indices = np.random.randint(0, labels.shape[0], 500000)

    labels = labels[rand_indices]
    preds = preds[rand_indices]

    err = np.abs(labels - preds)
    return err


def get_distribution(dataset: str):
    labels = np.load(f"./feat/{dataset}/agc_labels.npy").reshape(-1)

    rand_indices = np.random.randint(0, labels.shape[0], 500000)

    labels = labels[rand_indices]

    return labels


df = {dataset.split("/")[0]: get_err(dataset) for dataset in datasets}
df = pd.DataFrame.from_dict(df)
df_melted = df.melt(
    id_vars=[],
    var_name="dataset",
    value_name="number",
    value_vars=["pretrain", "coremof", "cof", "ppn", "zeolite"],
)
df_melted["type"] = np.array(["error" for _ in range(len(df_melted))])


df = {dataset.split("/")[0]: get_distribution(dataset) for dataset in datasets}
df = pd.DataFrame.from_dict(df)
dis_df_melted = df.melt(
    id_vars=[],
    var_name="dataset",
    value_name="number",
    value_vars=["pretrain", "coremof", "cof", "ppn", "zeolite"],
)
dis_df_melted["type"] = np.array(["distribution" for _ in range(len(dis_df_melted))])

result_df = pd.concat([df_melted, dis_df_melted], axis=0)

# df = pd.DataFrame(df)
palette = ["#cfcdcd", "#97a0b2", "#c69892", "#57746b", "#9e6f7a"]

sns.violinplot(
    data=result_df,
    x="dataset",
    y="number",
    hue="type",
    ax=ax,
    # inner="box",
    inner="quart",
    fill=True,
    palette=palette,
    linewidth=1,
    split=True,
    density_norm="width",
    bw_adjust=3,
    bw_method="silverman",
)
# sns.pointplot(data=[np.mean(result_df)], markers='o', color='white', linestyles='-', scale=0.8)
# sns.despine(left=True, bottom=True)
plt.ylim(-2.5, 8)
plt.xlabel(
    "Dataset", labelpad=10, fontdict={"size": 18, "family": "Arial", "color": "black"}
)
plt.ylabel(
    "Number", labelpad=10, fontdict={"size": 18, "family": "Arial", "color": "black"}
)
plt.title(
    "Error distribution for different porous materials",
    pad=20,
    fontdict={"size": 18, "family": "Arial", "color": "black"},
)
# ax.set_xticks([0, 1, 2, 3, 4])
ax.set_xticklabels(
    labels=[dataset.split("/")[0] for dataset in datasets],
    rotation=45,
    ha="right",
    fontdict={"size": 18, "family": "Arial", "color": "black"},
)
ax.set_yticks([i for i in range(-2, 9, 2)])
ax.set_yticklabels(
    labels=[i for i in range(-2, 9, 2)],
    # rotation=45,
    # ha="right",
    fontdict={"size": 18, "family": "Arial", "color": "black"},
)
# plt.savefig("./test.png", dpi=300, bbox_inches='tight')
plt.gca().spines["right"].set_visible(False)
plt.gca().spines["top"].set_visible(False)
plt.savefig("./png/transfer.png", format="png", dpi=300, bbox_inches="tight")
plt.savefig("./eps/transfer.eps", format="eps", dpi=300, bbox_inches="tight")



import matplotlib.pyplot as plt
import numpy as np
from scipy.stats import pearsonr
import seaborn as sns
from sklearn.metrics import r2_score
from einops import rearrange
import pandas as pd
import argparse

parser = argparse.ArgumentParser()
parser.add_argument("--dataset", type=str)
args = parser.parse_args()
dataset = args.dataset


def distance(pos1, pos2):
    def get_pos(a):
        az = a % 10
        a = a // 10
        ay = a % 10
        a = a // 10
        ax = a % 10
        a = a // 10
        return ax, ay, az

    ss = np.sum((np.array(get_pos(pos1)) - np.array(get_pos(pos2))) ** 2)
    # dis = np.sqrt(ss)

    # if ss == 25 or ss == 50 or ss == 49:
    #     return dis
    # else:
    #     return 0

    return ss


show = np.load(f"./feat/{dataset}/self_attns.npy")
print("Load end, attn shape: ", show.shape)
# show = np.mean(show, axis=0)

# show = rearrange(show.reshape((10, 10, 10, -1)), "z y x a -> x y z a")
show = rearrange(show.reshape((-1, 10, 10, 10, 1000)), "b z y x a -> b x y z a")


# Scatter plot

distance_matrix = np.array([[distance(i, j) for j in range(1000)] for i in range(1000)])
# uniq_distance = np.unique(distance_matrix)
# sort_distance = np.sort(uniq_distance)

print("Distance matrix generate end")

rand_indices = np.random.randint(0, show.shape[1], size=30)
show = show[rand_indices]
distance_matrix = np.tile(distance_matrix, [len(rand_indices), 1, 1])

print("Sample end", show.shape, distance_matrix.shape)

show = show.reshape(-1)
distance_matrix = distance_matrix.reshape(-1)

# for dis in sort_distance:
#     mean_attn = np.mean(show[np.where(distance_matrix == dis)])
#     print(f"Dis: {dis}, attn: {mean_attn}")

# rand_indices = np.where(distance_matrix < 150)
# show = show[rand_indices]
# distance_matrix = distance_matrix[rand_indices]

# rand_indices = np.where(show > 0.01)
# show = show[rand_indices]
# distance_matrix = distance_matrix[rand_indices]

rand_indices = np.random.randint(0, show.shape[0], size=2000000)
show = show[rand_indices]
distance_matrix = distance_matrix[rand_indices]

df = {"attention": show, "distance": distance_matrix}
df = pd.DataFrame(df)

# sns.scatterplot(x=distance_matrix, y=show, s=5)

# plt.savefig("./scatter.png", dpi=300, format="png")
# plt.savefig("./scatter.eps", dpi=600, format="eps")

# sns.kdeplot(x=distance_matrix, y=show, cmap="Blues", fill=True)
# plt.savefig("./kde.png", dpi=300, format="png")
# plt.savefig("./kde.eps", dpi=600, format="eps")

g = sns.lineplot(x=distance_matrix, y=show, errorbar="pi", color="#8389a0")
g.set(ylim=[0, 0.06])
g.set(xlim=[0, 100])
sns.despine(offset=10, trim=True)
# plt.xlim([0, 100])
# plt.axis('auto')
plt.savefig("./png/self_attn.lineplot.png", dpi=300, format="png")
plt.savefig("./eps/self_attn.lineplot.eps", dpi=600, format="eps")

# sns.lmplot(data=df, x="distance", y="attention")

print(r2_score(distance_matrix, show))



import pandas as pd
from pathlib import Path
import pickle
import numpy as np
import argparse
from tqdm import tqdm
from spbnet.utils.echo import start, end, title, err, param


def checkData(root_dir, modal_dir, prop_dir, prop_name, split):

    root_dir = Path(root_dir)
    modal_dir = root_dir / modal_dir
    prop_dir = root_dir / prop_dir

    def check_cifid(cifid: str):
        try:
            GRID = 30
            channel = 20

            data_dir = modal_dir
            graphdata = pickle.load(
                (modal_dir / "graphdata" / f"{cifid}.graphdata").open("rb")
            )
            # graphdata = pickle.load(open(f"./spbnet/graphdata/{cifid}.graphdata", "rb"))
            file_grid = data_dir / "grid" / f"{cifid}.grid"
            file_griddata = data_dir / "griddata8" / f"{cifid}.npy"

            # get grid
            with file_grid.open("r") as f:
                lines = f.readlines()
                a, b, c = [float(i) for i in lines[0].split()[1:]]
                angle_a, angle_b, angle_c = [float(i) for i in lines[1].split()[1:]]
                cell = [int(i) for i in lines[2].split()[1:]]

            griddata = np.load(file_griddata.absolute())  # [30*30*30, 20] uint8
            griddata = griddata.astype(np.float32)

            griddata = griddata.reshape(GRID, GRID, GRID, channel)

            return None
        except Exception as e:
            return e

    title("CHECKING MODAL DATA")
    param(modal_dir=modal_dir.absolute(), prop_name=prop_name)

    def check_idprop(idprop_path):
        start(f"Checking {idprop_path.absolute()}")
        df = pd.read_csv(idprop_path.absolute())

        for item in tqdm(df.iloc, total=len(df)):
            cifid = item["cifid"]
            res = check_cifid(cifid)
            if res is not None:
                err(f"Error when checing {cifid}: {res}")
                return
        end(f"Check {idprop_path.absolute()} success")

    if split == "all":
        for split in ["train", "test", "validate"]:
            idprop_path = prop_dir / f"{prop_name}.{split}.csv"
            check_idprop(idprop_path=idprop_path)
    elif split is None:
        idprop_path = prop_dir / f"{prop_name}.csv"
        check_idprop(idprop_path)
    else:
        idprop_path = prop_dir / f"{prop_name}.{split}.csv"
        check_idprop(idprop_path)

    title("CHECK MODAL DATA END")



from typing import Any, Optional
from functools import partial

# from pytorch_lightning.utilities.types import STEP_OUTPUT
# from model.crossformer import CrossFormer
from spbnet.modules.module import CrossFormer
from spbnet.modules.heads import RegressionHead, Pooler, ClassificationHead
from spbnet.modules.optimize import set_scheduler
from spbnet.modules import objectives
from torch import nn
from spbnet.data.dataset_feat import Dataset
import pytorch_lightning as pl
import pandas as pd
import json
import yaml
import torch
from pytorch_lightning.callbacks import ModelCheckpoint
from pytorch_lightning.callbacks import LearningRateMonitor
from torch.utils.data import DataLoader
from sklearn.model_selection import train_test_split
from pathlib import Path
from multiprocessing import Process
from pytorch_lightning.loggers import TensorBoardLogger
from einops import rearrange
from torch import optim
import numpy as np
from spbnet.utils.echo import title, start, end, param, err

cur_dir = Path(__file__).parent


def r2_score(y_true, y_pred):
    # 计算总平均值
    mean_true = torch.mean(y_true)

    # 计算总平方和
    total_ss = torch.sum((y_true - mean_true) ** 2)

    # 计算残差平方和
    resid_ss = torch.sum((y_true - y_pred) ** 2)

    # 计算R2分数
    r2 = 1 - (resid_ss / total_ss)

    return r2


class CrossFormerTrainer(pl.LightningModule):
    def __init__(self, config: dict):
        super(CrossFormerTrainer, self).__init__()
        self.save_hyperparameters()

        self.config = config

        self.model_config = config
        self.optimizer_config = config

        model_config = self.model_config
        self.model = CrossFormer(model_config)
        # print(self.model)
        # self.model = AtomFormer(config)

        self.mean = config["mean"]
        self.std = config["std"]

        # pooler
        self.pooler = Pooler(model_config["hid_dim"])
        self.pooler.apply(objectives.init_weights)

        # void fraction prediction
        self.vfp_head = RegressionHead(model_config["hid_dim"])
        self.vfp_head.apply(objectives.init_weights)

        # topo classify
        self.tc_head = ClassificationHead(model_config["hid_dim"], config["topo_num"])
        self.tc_head.apply(objectives.init_weights)

        # atom grid classify
        self.agc_head = nn.Linear(model_config["hid_dim"], 1)
        self.agc_head.apply(objectives.init_weights)

        self.mse_loss = nn.MSELoss()
        self.mae_loss = nn.L1Loss()

        # mae
        self.min_mae = 1e5

    def on_test_epoch_start(self) -> None:
        self.cifids = []

        self.feats = []
        self.structure_feats = []
        self.potential_feats = []
        self.agc_preds = []
        self.agc_labels = []
        self.self_attns = []

        return super().on_test_epoch_start()

    def test_step(self, batch, batch_idx) -> torch.Tensor:
        feat = self.model(batch)
        self.cifids += feat["cifid"]
        batch_size = len(feat["cifid"])

        # item
        cls_feat = feat["cls_feat"]
        cls_feat = self.pooler(cls_feat).detach().cpu().numpy()
        structure_feat = (
            feat["structure_feat"].detach().cpu().numpy()
        )  # [B, max_graph_len, hid_dim]
        structure_mask = (
            feat["structure_mask"].detach().cpu().numpy()
        )  # [B, max_graph_len]
        potential_feat = feat["potential_feat"]  # [B, 1004, hid_dim]
        agc_pred = (
            self.agc_head(potential_feat)
            .detach()
            .cpu()
            .numpy()
            .reshape((batch_size, -1))
        )  # [B, 1004]
        potential_feat = potential_feat.detach().cpu().numpy()
        atomgrid: np.array = batch["atomgrid"].detach().cpu().numpy()
        agc_label = atomgrid
        patch_size = self.model_config["patch_size"]["lj"]
        agc_label = rearrange(
            agc_label,
            "b (h p1) (w p2) (d p3) -> b (h w d) (p1 p2 p3)",
            p1=patch_size,
            p2=patch_size,
            p3=patch_size,
        )
        agc_label = np.sum(
            agc_label, axis=-1
        )  # [B, GRID / PatchSize * GRID / PatchSize * GRID / PatchSize]
        self_attn = feat["sa_attn"].detach().cpu().numpy()  # [B, 1004, 1004]
        cross_attn = feat["ca_attn"].detach().cpu().numpy()  # [B, 1004, max_graph_len]

        # sample
        # structure
        structure_feat = structure_feat.reshape(
            (-1, structure_feat.shape[-1])
        )  # [B * max_graph_len, hid_dim]
        structure_mask = structure_mask.reshape(-1)  # [B * max_graph_len]
        indices_of_ones = np.where(structure_mask == 0)[0]
        structure_indices = np.random.choice(
            indices_of_ones, size=20 * batch_size, replace=False
        )
        structure_feat = structure_feat[structure_indices]  # [20 * B, hid_dim]

        # potential
        potential_feat = potential_feat.reshape((-1, potential_feat.shape[-1]))
        agc_pred = agc_pred.reshape(-1)
        agc_label = agc_label.reshape(-1)
        potential_indices = np.random.randint(
            0, potential_feat.shape[0], 20 * batch_size
        )
        potential_feat = potential_feat[potential_indices]
        agc_pred = agc_pred[potential_indices]
        agc_label = agc_label[potential_indices]

        # attn
        attn_indices = np.random.randint(0, batch_size, 5)
        self_attn = self_attn[attn_indices, 2:-2, 2:-2]  # [5, 1000, 1000]

        for feat_name in self.config["feats"]:
            if feat_name == "potential":
                self.potential_feats.append(potential_feat)  # [20 * B, hid_dim]
            if feat_name == "structure":
                self.structure_feats.append(structure_feat)  # [20 * B, hid_dim]
            if feat_name == "agc_pred":
                self.agc_preds.append(agc_pred)  # [20 * B]
            if feat_name == "agc_label":
                self.agc_labels.append(agc_label)  # [20 * B]
            if feat_name == "feat":
                self.feats.append(feat)  # [B]
            if feat_name == "self_attn":
                self.self_attns.append(self_attn)  # [5, 1000, 1000]

    def on_test_epoch_end(self) -> None:
        save_dir = Path(self.config["save_dir"], parents=True)
        save_dir.mkdir(exist_ok=True)

        for feat_name in self.config["feats"]:
            if feat_name == "potential":
                potential_feats = np.concatenate(self.potential_feats)
                np.save(
                    save_dir / f"potential_feats.npy", potential_feats
                )  # [N * 20, hid_dim]
            if feat_name == "structure":
                structure_feats = np.concatenate(self.structure_feats)
                np.save(
                    save_dir / f"structure_feats.npy", structure_feats
                )  # [N * 20, hid_dim]
            if feat_name == "agc_pred":
                agc_preds = np.concatenate(self.agc_preds)
                np.save(save_dir / f"agc_preds.npy", agc_preds)  # [N * 20]
            if feat_name == "agc_label":
                agc_labels = np.concatenate(self.agc_labels)
                np.save(save_dir / f"agc_labels.npy", agc_labels)  # [N * 20]
            if feat_name == "feat":
                feats = np.concatenate(self.feats)
                np.save(save_dir / f"feats.npy", feats)  # [N, hid_dim]
            if feat_name == "self_attn":
                self_attns = np.concatenate(self.self_attns)
                np.save(
                    save_dir / f"self_attns.npy", self_attns
                )  # [N * 5 / B, 1000, 1000]
        return super().on_test_epoch_end()


def feat(config_path: str):
    title("START TO OBTAIN FEATURES")

    base_config_path = (cur_dir / "config.yaml").absolute()
    with open(base_config_path, "r") as f:
        base_config = yaml.load(f, Loader=yaml.FullLoader)

    with open(config_path, "r") as f:
        user_config = yaml.load(f, Loader=yaml.FullLoader)

    if user_config.get("data_dir") is None:
        err(f"Please specify modal directory `data_dir`!")
        return
    if user_config.get("id_prop") is None:
        err(f"Please specify label data `id_prop`")
        return
    if user_config.get("task") is None:
        err(f"Please specify task")
        return
    if user_config.get("log_dir") is None:
        err(f"Please specify log directory")
        return
    if user_config.get("save_dir") is None:
        err(f"Please specify save directory")
        return

    base_config.update(user_config)
    config = base_config

    # check
    device = config["device"]
    task = config["task"]

    # handle
    id_prop_path = Path(config["id_prop"])
    ckpt_path = Path(config["ckpt"])  # TODO: CKPT
    data_dir = Path(config["data_dir"])
    log_dir = Path(config["log_dir"])

    # split train & val
    id_prop_dir = id_prop_path.parent
    test_df = pd.read_csv(id_prop_path.absolute(), dtype={"cifid": str})
    test_dataset = Dataset(test_df, data_dir, config["nbr_fea_len"], task)

    # id_prop
    id_prop_df = test_df.copy()
    filter_id_prop_df = id_prop_df.dropna(subset=[task])

    # mean & std
    mean = None
    std = None
    mean = filter_id_prop_df[task].mean()
    std = filter_id_prop_df[task].std()
    config["mean"] = float(mean)
    config["std"] = float(std)

    # train
    checkpoint_callback = ModelCheckpoint(monitor="val_mae", mode="min", save_last=True)
    lr_monitor = LearningRateMonitor(logging_interval="epoch")
    logger = TensorBoardLogger(save_dir=(log_dir / task).absolute(), name="")
    trainer = pl.Trainer(
        max_epochs=config["epochs"],
        min_epochs=0,
        devices=device,
        accelerator=config["accelerator"],
        strategy=config["strategy"],
        callbacks=[checkpoint_callback, lr_monitor],
        accumulate_grad_batches=config["accumulate_grad_batches"],
        precision=config["precision"],
        log_every_n_steps=config["log_every_n_steps"],
        logger=logger,
    )
    model = CrossFormerTrainer(config)
    if not config["ckpt"] == "scratch":
        ckpt = torch.load(ckpt_path)
        model.load_state_dict(ckpt["state_dict"], strict=True)

    trainer.test(
        model=model,
        dataloaders=DataLoader(
            test_dataset,
            batch_size=config["batch_size"],
            shuffle=False,
            num_workers=4,
            collate_fn=partial(Dataset.collate, img_size=config["img_size"]),
        ),
        ckpt_path=None,
    )


def main():
    config_path = "./config.yaml"
    with open(config_path, "r") as f:
        config = yaml.load(f, Loader=yaml.FullLoader)
    tasks = config["tasks"]

    test(0, config_path)


if __name__ == "__main__":
    main()



from pathlib import Path
import math
import numpy as np
from .prepare_data import _make_supercell
from ase.io import read
import argparse
from spbnet.utils.echo import title, param, err
from multiprocessing import Process
from tqdm import tqdm


GRID = 30


def get_atoms(cif_path: Path):
    atoms = read(cif_path.absolute())
    atoms = _make_supercell(atoms, cutoff=30)
    return atoms


def cif2array(cif_path: Path):
    ret = np.zeros((GRID, GRID, GRID))

    atoms = get_atoms(
        cif_path=cif_path,
    )
    frac_positions = atoms.get_scaled_positions()
    atomic_numbers = atoms.get_atomic_numbers()

    # 计算小格子中每个原子的个数
    for frac_coords, _ in zip(frac_positions, atomic_numbers):
        frac_ints = [math.floor(frac_coords[i] * GRID) for i in range(3)]
        ret[frac_ints[0], frac_ints[1], frac_ints[2]] += 1

    return ret


def buildAtomGridi(cif_path: Path, tgt_dir: Path):
    (tgt_dir / "atomgrid").mkdir(exist_ok=True)
    array = cif2array(cif_path)  # [GRID, GRID, GRID]
    np.save((tgt_dir / "atomgrid" / f"{cif_path.stem}").absolute(), array)


def process_multi_cif(cif_paths, target_dir: Path):
    for cif_path in tqdm(cif_paths):
        try:
            buildAtomGridi(cif_path, target_dir)
        except Exception as e:
            err(f"Error when processing {cif_path.stem}: {e}")


def buildAtomGrid(
    root_dir: str, cif_dir: str = "cif", target_dir: str = "spbnet", n_process: int = 1
):

    root_dir = Path(".") / root_dir
    cif_dir = root_dir / cif_dir
    target_dir = root_dir / target_dir

    cif_paths = list(cif_dir.iterdir())
    cif_paths = list(filter(lambda item: item.suffix == ".cif", cif_paths))
    # cif_paths = cif_paths[:100]

    title("BUILDING ATOM GRID")
    param(
        cif_dir=cif_dir.absolute(),
        target_dir=target_dir.absolute(),
        cif_num=len(cif_paths),
    )

    process_num = n_process

    cif_num_per_process = len(cif_paths) // process_num + 1
    processes = []
    for i in range(process_num):
        cif_paths_i = cif_paths[i * cif_num_per_process : (i + 1) * cif_num_per_process]
        process = Process(target=process_multi_cif, args=(cif_paths_i, target_dir))
        process.start()
        processes.append(process)
    for process in processes:
        process.join()

    # for cif_path in tqdm(list(target_dir.iterdir())):
    #     buildAtomGridi(cif_path, target_dir)
    title("BUILD ATOM GRID END")

    return



import pickle
import pandas as pd
from pathlib import Path

import numpy as np

import torch


class Dataset(torch.utils.data.Dataset):
    def __init__(
        self,
        df: pd.DataFrame,
        data_dir: Path,
        nbr_fea_len: int,
        task: str,
        draw_false_grid=False,
    ):
        super().__init__()
        filter_df = df.dropna(subset=[task])
        self.df = filter_df
        self.task = task
        self.data_dir = data_dir
        self.draw_false_grid = draw_false_grid
        self.nbr_fea_len = nbr_fea_len

        self.GRID = 30

        self.max_atom_num = 512

    def __len__(self):
        return len(self.df)

    # grid
    @staticmethod
    def handle_griddata_float(
        file_griddata, emin=-5000.0, emax=5000, bins=101, GRID=30, channel=20
    ):
        griddata = np.fromfile(file_griddata, dtype=np.float32)
        # griddata = griddata.astype(np.float16)  # [30 * 30 * 30 * 20]
        griddata = torch.from_numpy(griddata)
        griddata = griddata / 250 * 101 + 1

        griddata = griddata.reshape(GRID, GRID, GRID, channel)
        # lj = griddata[:, :, :, 12] - griddata[:, :, :, 19]
        lj = griddata[:, :, :, 12]
        correction = torch.cat(
            [griddata[:, :, :, :12], griddata[:, :, :, 13:19]], dim=-1
        )

        # griddata[griddata > 6e4] = 6e4
        # griddata[griddata < -6e4] = -6e4

        lj[lj <= emin] = emin
        lj[lj > emax] = emax

        correction[correction <= emin] = emin
        correction[correction > emax] = emax

        # x = np.linspace(emin, emax, bins)
        # lj = np.digitize(lj, x) + 1
        # correction = np.digitize(correction, x) + 1

        return lj, correction

    @staticmethod
    def handle_griddata(file_griddata: Path, GRID=30, channel=20):
        griddata = np.load(file_griddata.absolute())  # [30*30*30, 20] uint8
        griddata = griddata.astype(np.float32)

        griddata = griddata.reshape(GRID, GRID, GRID, channel)
        lj = griddata[:, :, :, 12]
        correction = np.concatenate(
            [griddata[:, :, :, :12], griddata[:, :, :, 13:19]], axis=-1
        )

        return torch.from_numpy(lj).float(), torch.from_numpy(correction).float()

    @staticmethod
    def calculate_volume(a, b, c, angle_a, angle_b, angle_c):
        a_ = np.cos(angle_a * np.pi / 180)
        b_ = np.cos(angle_b * np.pi / 180)
        c_ = np.cos(angle_c * np.pi / 180)

        v = a * b * c * np.sqrt(np.abs(1 - a_**2 - b_**2 - c_**2 + 2 * a_ * b_ * c_))

        return v.item() / (60 * 60 * 60)  # normalized volume

    def get_raw_grid_data(self, cifid):
        file_grid = self.data_dir / "grid" / f"{cifid}.grid"
        file_griddata = self.data_dir / "griddata8" / f"{cifid}.npy"

        # get grid
        with file_grid.open("r") as f:
            lines = f.readlines()
            a, b, c = [float(i) for i in lines[0].split()[1:]]
            angle_a, angle_b, angle_c = [float(i) for i in lines[1].split()[1:]]
            cell = [int(i) for i in lines[2].split()[1:]]

        volume = self.calculate_volume(a, b, c, angle_a, angle_b, angle_c)

        # get grid data
        lj, corr = self.handle_griddata(file_griddata)

        return cell, volume, lj, corr

    def get_grid_data(self, cifid, draw_false_grid=False):
        cell, volume, lj, corr = self.get_raw_grid_data(cifid)
        return {"cell": cell, "volume": volume, "lj": lj, "corr": corr}

    @staticmethod
    def get_gaussian_distance(distances, num_step, dmax, dmin=0, var=0.2):
        """
        Expands the distance by Gaussian basis
        (https://github.com/txie-93/cgcnn.git)
        """

        assert dmin < dmax
        _filter = np.linspace(
            dmin, dmax, num_step
        )  # = np.arange(dmin, dmax + step, step) with step = 0.2

        return np.exp(-((distances[..., np.newaxis] - _filter) ** 2) / var**2).float()

    # @lru_cache(maxsize=None)
    def get_graph(self, cifid: str):
        # moftransformer
        file_graph = self.data_dir / "graphdata" / f"{cifid}.graphdata"

        graphdata = pickle.load(file_graph.open("rb"))
        # graphdata = ["cifid", "atom_num", "nbr_idx", "nbr_dist", "uni_idx", "uni_count"]
        atom_num = torch.LongTensor(graphdata[1].copy())
        nbr_idx = torch.LongTensor(graphdata[2].copy()).view(len(atom_num), -1)
        nbr_dist = torch.FloatTensor(graphdata[3].copy()).view(len(atom_num), -1)

        nbr_fea = torch.FloatTensor(
            self.get_gaussian_distance(nbr_dist, num_step=self.nbr_fea_len, dmax=8)
        )

        uni_idx = graphdata[4]
        uni_count = graphdata[5]

        return {
            "atom_num": atom_num,
            "nbr_idx": nbr_idx,
            "nbr_fea": nbr_fea,
            "uni_idx": uni_idx,
            "uni_count": uni_count,
        }

    def get_coulomb(self, cifid: str):
        file_coulomb = self.data_dir / "coulomb" / f"{cifid}.coulomb"
        potential = torch.load(file_coulomb, map_location="cpu")
        return {"coulomb": potential}

    def __getitem__(self, index):
        ret = dict()

        item = self.df.iloc[index]
        cifid = item["cifid"]
        target = float(item[self.task])

        ret.update(
            {
                "cifid": cifid,
                "target": target,
            }
        )
        # ret.update(self.get_atom(cifid))
        ret.update(self.get_grid_data(cifid, draw_false_grid=self.draw_false_grid))
        ret.update(self.get_graph(cifid))
        # ret.update(self.get_coulomb(cifid))

        # ret.update(self.get_tasks(index))

        return ret

    @staticmethod
    def collate(batch, img_size):
        """
        collate batch
        Args:
            batch (dict): [cifid, atom_num, nbr_idx, nbr_fea, uni_idx, uni_count,
                            grid_data, cell, (false_grid_data, false_cell), target]
            img_size (int): maximum length of img size

        Returns:
            dict_batch (dict): [cifid, atom_num, nbr_idx, nbr_fea, crystal_atom_idx,
                                uni_idx, uni_count, grid, false_grid_data, target]
        """
        batch_size = len(batch)
        keys = set([key for b in batch for key in b.keys()])

        dict_batch = {k: [dic[k] if k in dic else None for dic in batch] for k in keys}

        # graph
        # moftransformer
        batch_atom_num = dict_batch["atom_num"]
        batch_nbr_idx = dict_batch["nbr_idx"]
        batch_nbr_fea = dict_batch["nbr_fea"]

        crystal_atom_idx = []
        base_idx = 0
        for i, nbr_idx in enumerate(batch_nbr_idx):
            n_i = nbr_idx.shape[0]
            crystal_atom_idx.append(torch.arange(n_i) + base_idx)
            nbr_idx += base_idx
            base_idx += n_i

        dict_batch["atom_num"] = torch.cat(batch_atom_num, dim=0)
        dict_batch["nbr_idx"] = torch.cat(batch_nbr_idx, dim=0)
        dict_batch["nbr_fea"] = torch.cat(batch_nbr_fea, dim=0)
        dict_batch["crystal_atom_idx"] = crystal_atom_idx

        # target
        batch_target = dict_batch["target"]
        batch_target = torch.tensor(batch_target, dtype=torch.float)
        dict_batch["target"] = batch_target

        # # coulomb
        # batch_coulomb = dict_batch["coulomb"]  # List[[H, W, D]]
        # batch_coulomb = torch.stack(batch_coulomb)  # [B, H, W, D]
        # # batch_coulomb = torch.clamp(batch_coulomb, min=None, max=300)
        # # batch_coulomb = normalize(batch_coulomb)
        # dict_batch["coulomb"] = batch_coulomb

        # grid
        batch_lj_data = dict_batch["lj"]
        batch_lj_data = torch.stack(batch_lj_data)
        batch_lj_data = batch_lj_data.view(batch_lj_data.shape[0], 30, 30, 30)
        dict_batch["lj"] = batch_lj_data

        batch_corr_data = dict_batch["corr"]
        batch_corr_data = torch.stack(batch_corr_data)
        batch_corr_data = batch_corr_data.view(batch_corr_data.shape[0], 30, 30, 30, 18)
        dict_batch["corr"] = batch_corr_data

        return dict_batch



from spbnet.modules.module import CrossFormer
import yaml
from pathlib import Path
from ase.io import read
import math
import numpy as np
from ase.build import make_supercell
import matplotlib.pyplot as plt
from matplotlib.colors import Normalize
import torch
import pickle
from ase.io import write
from pathlib import Path
import os
from ase.visualize.plot import plot_atoms


def get_grid_data(data_dir, cifid):
    def handle_griddata(file_griddata: Path, GRID=30, channel=20):
        griddata = np.load(file_griddata.absolute())  # [30*30*30, 20] uint8
        griddata = griddata.astype(np.float32)

        griddata = griddata.reshape(GRID, GRID, GRID, channel)
        # lj = griddata[:, :, :, 12] - griddata[:, :, :, 19]
        lj = griddata[:, :, :, 12]
        # ls = np.where(lj > 100, 0, lj)
        correction = np.concatenate(
            [griddata[:, :, :, :12], griddata[:, :, :, 13:19]], axis=-1
        )

        return torch.from_numpy(lj).float(), torch.from_numpy(correction).float()

    def calculate_volume(a, b, c, angle_a, angle_b, angle_c):
        a_ = np.cos(angle_a * np.pi / 180)
        b_ = np.cos(angle_b * np.pi / 180)
        c_ = np.cos(angle_c * np.pi / 180)

        v = a * b * c * np.sqrt(1 - a_**2 - b_**2 - c_**2 + 2 * a_ * b_ * c_)

        return v.item() / (60 * 60 * 60)  # normalized volume

    def get_raw_grid_data(cifid):
        file_grid = data_dir / "grid" / f"{cifid}.grid"
        file_griddata = data_dir / "griddata8" / f"{cifid}.npy"

        # get grid
        with file_grid.open("r") as f:
            lines = f.readlines()
            a, b, c = [float(i) for i in lines[0].split()[1:]]
            angle_a, angle_b, angle_c = [float(i) for i in lines[1].split()[1:]]
            cell = [int(i) for i in lines[2].split()[1:]]

        volume = calculate_volume(a, b, c, angle_a, angle_b, angle_c)

        # get grid data
        lj, corr = handle_griddata(file_griddata)

        return cell, volume, lj, corr

    def get_grid_data(cifid):
        cell, volume, lj, corr = get_raw_grid_data(cifid)
        return {"cell": cell, "volume": volume, "lj": lj, "corr": corr}

    return get_grid_data(cifid)


def get_graph(data_dir, cifid: str, nbr_fea_len=64):
    def get_gaussian_distance(distances, num_step, dmax, dmin=0, var=0.2):
        """
        Expands the distance by Gaussian basis
        (https://github.com/txie-93/cgcnn.git)
        """

        assert dmin < dmax
        _filter = np.linspace(
            dmin, dmax, num_step
        )  # = np.arange(dmin, dmax + step, step) with step = 0.2

        return np.exp(-((distances[..., np.newaxis] - _filter) ** 2) / var**2).float()

    # moftransformer
    file_graph = data_dir / "graphdata" / f"{cifid}.graphdata"

    graphdata = pickle.load(file_graph.open("rb"))
    # graphdata = ["cifid", "atom_num", "nbr_idx", "nbr_dist", "uni_idx", "uni_count"]
    atom_num = torch.LongTensor(graphdata[1].copy())
    nbr_idx = torch.LongTensor(graphdata[2].copy()).view(len(atom_num), -1)
    nbr_dist = torch.FloatTensor(graphdata[3].copy()).view(len(atom_num), -1)

    nbr_fea = torch.FloatTensor(
        get_gaussian_distance(nbr_dist, num_step=nbr_fea_len, dmax=8)
    )

    uni_idx = graphdata[4]
    uni_count = graphdata[5]

    return {
        "atom_num": atom_num,
        "nbr_idx": nbr_idx,
        "nbr_fea": nbr_fea,
        "uni_idx": uni_idx,
        "uni_count": uni_count,
    }


def collate(batch):
    """
    collate batch
    Args:
        batch (dict): [cifid, atom_num, nbr_idx, nbr_fea, uni_idx, uni_count,
                        grid_data, cell, (false_grid_data, false_cell), target]
        img_size (int): maximum length of img size

    Returns:
        dict_batch (dict): [cifid, atom_num, nbr_idx, nbr_fea, crystal_atom_idx,
                            uni_idx, uni_count, grid, false_grid_data, target]
    """
    batch_size = len(batch)
    keys = set([key for b in batch for key in b.keys()])

    dict_batch = {k: [dic[k] if k in dic else None for dic in batch] for k in keys}

    # graph
    # moftransformer
    batch_atom_num = dict_batch["atom_num"]
    batch_nbr_idx = dict_batch["nbr_idx"]
    batch_nbr_fea = dict_batch["nbr_fea"]

    crystal_atom_idx = []
    base_idx = 0
    for i, nbr_idx in enumerate(batch_nbr_idx):
        n_i = nbr_idx.shape[0]
        crystal_atom_idx.append(torch.arange(n_i) + base_idx)
        nbr_idx += base_idx
        base_idx += n_i

    dict_batch["atom_num"] = torch.cat(batch_atom_num, dim=0)
    dict_batch["nbr_idx"] = torch.cat(batch_nbr_idx, dim=0)
    dict_batch["nbr_fea"] = torch.cat(batch_nbr_fea, dim=0)
    dict_batch["crystal_atom_idx"] = crystal_atom_idx

    # target
    batch_target = dict_batch["target"]
    batch_target = torch.tensor(batch_target, dtype=torch.float)
    dict_batch["target"] = batch_target

    # grid
    batch_lj_data = dict_batch["lj"]
    batch_lj_data = torch.stack(batch_lj_data)
    batch_lj_data = batch_lj_data.view(batch_lj_data.shape[0], 30, 30, 30)
    dict_batch["lj"] = batch_lj_data

    batch_corr_data = dict_batch["corr"]
    batch_corr_data = torch.stack(batch_corr_data)
    batch_corr_data = batch_corr_data.view(batch_corr_data.shape[0], 30, 30, 30, 18)
    dict_batch["corr"] = batch_corr_data

    return dict_batch


def get_atoms(cif_file: Path, cutoff=8):
    def _make_supercell(atoms, cutoff):
        """
        make atoms into supercell when cell length is less than cufoff (min_length)
        """
        # when the cell lengths are smaller than radius, make supercell to be longer than the radius
        scale_abc = []
        for l in atoms.cell.cellpar()[:3]:
            if l < cutoff:
                scale_abc.append(math.ceil(cutoff / l))
            else:
                scale_abc.append(1)

        # make supercell
        m = np.zeros([3, 3])
        np.fill_diagonal(m, scale_abc)
        atoms = make_supercell(atoms, m)
        return atoms

    atoms = read(str(cif_file.absolute()))
    atoms = _make_supercell(atoms, cutoff=cutoff)
    return atoms



# MOFTransformer version 2.0.0
class CLICommand:
    """
    Make GRIDAY which calculated potential basis function.
    """

    @staticmethod
    def add_arguments(parser):
        pass

    @staticmethod
    def run(args):
        from spbnet.data.install_griday import make_griday

        make_griday()



import numpy as np
from scipy.stats import pearsonr
from sklearn.metrics import r2_score

import seaborn as sns
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib as mpl
from spbnet.utils.echo import start, end


def boxplot(agc_labels: np.array, agc_preds: np.array):
    start(
        f"Load end: label({agc_labels.shape}), pred({agc_preds.shape}), start to draw box plot"
    )

    print("R2", r2_score(agc_labels, agc_preds))
    print("pearsonr", pearsonr(agc_labels, agc_preds)[0])
    idxes = np.where(agc_labels > 0)[0]
    agc_labels = agc_labels[idxes]
    agc_preds = agc_preds[idxes]

    idxes = np.where(agc_labels < 10)[0]
    agc_labels = agc_labels[idxes]
    agc_preds = agc_preds[idxes]

    data = {"agc_labels": agc_labels, "agc_preds": agc_preds}

    data = pd.DataFrame(data)

    mpl.rcParams["font.size"] = 14

    fig = plt.figure(figsize=(9, 7))

    subfig, ax = plt.subplots(figsize=(8, 6))
    sns.boxplot(
        x="agc_labels",
        y="agc_preds",
        data=data,
        showfliers=False,
        ax=ax,
        color="#d8bfca",
    )
    sns.despine(offset=10, trim=True)
    plt.xlabel(
        "Actual atom number",
        labelpad=10,
        fontdict={"size": 18, "family": "Arial", "color": "black"},
    )
    plt.ylabel(
        "Predicted atom number",
        labelpad=10,
        fontdict={"size": 18, "family": "Arial", "color": "black"},
    )
    plt.title(
        "Boxplot of predicted atom number",
        pad=20,
        fontdict={"size": 18, "family": "Arial", "color": "black"},
    )

    # plt.savefig("./eps/agc.boxplot.eps", format="eps", bbox_inches="tight")
    plt.savefig("./agc.boxplot.png", format="png", bbox_inches="tight")
    end(f"draw end, file saved in ./agc.boxplot.png")


def jointplot(agc_labels, agc_preds):
    start(
        f"Load end: label({agc_labels.shape}), pred({agc_preds.shape}), start to line plot"
    )

    print("R2", r2_score(agc_labels, agc_preds))
    print("pearsonr", pearsonr(agc_labels, agc_preds)[0])

    idxes = np.where(agc_labels > 0)[0]
    agc_labels = agc_labels[idxes]
    agc_preds = agc_preds[idxes]

    idxes = np.where(agc_labels < 10)[0]
    agc_labels = agc_labels[idxes]
    agc_preds = agc_preds[idxes]

    data = {"agc_labels": agc_labels, "agc_preds": agc_preds}

    data = pd.DataFrame(data)

    mpl.rcParams["font.size"] = 14

    fig = plt.figure(figsize=(9, 7))

    subfig, ax = plt.subplots(figsize=(8, 6))

    sns.lineplot(
        x=agc_labels,
        y=agc_preds,
    )
    sns.jointplot(x=agc_labels, y=agc_preds, kind="hex", color="#c08292")
    sns.despine(offset=10, trim=True)

    plt.xlabel(
        "Actual number of atoms", labelpad=10, fontdict={"size": 16, "family": "Arial"}
    )
    plt.ylabel(
        "Predicted number of atoms",
        labelpad=10,
        fontdict={"size": 16, "family": "Arial"},
    )

    # plt.savefig("./eps/agc.jointplot.eps", format="eps", bbox_inches="tight")
    plt.savefig("./agc.jointplot.png", dpi=600, format="png", bbox_inches="tight")

    end(f"draw end, file saved in ./agc.jointplot.png")


# Rotate the starting point around the cubehelix hue circle
def kde(
    data: np.array,  # [N, N]
    emin=0,
    emax=10,
    lowdark=True,
    filename="agc.kde",
    bw_adjust=0.43,
    figsize=(6, 6),
):
    # Set up the matplotlib figure
    # f, axes = plt.subplots(3, 3, figsize=(9, 9), sharex=True, sharey=True)
    fig = plt.figure(figsize=figsize)
    # ax = fig.add_subplot()
    ax = fig.add_axes((0.0, 0.0, 1.0, 1.0))
    # width_px, height_px = (
    #     fig.get_size_inches() * fig.dpi
    # )  # 获取宽度和高度（以像素为单位）
    # print(f"Width: {width_px}, Height: {height_px}")

    x = []
    y = []

    data = np.clip(data, emin, emax)

    # 归一化处理
    data = (data - emin) / (emax - emin)
    for i in range(data.shape[0]):
        for j in range(data.shape[1]):
            value = data[i][j]
            if lowdark:
                num = 10 - int(value * 10)
            else:
                num = int(value * 2)
            for _ in range(num):
                x.append(i / 15 - 1)
                y.append(j / 15 - 1)

    # # Create a cubehelix colormap to use with kdeplot
    # cmap = sns.cubehelix_palette(start=s, light=1, as_cmap=True)

    # Generate and plot a random bivariate dataset
    # x, y = rs.normal(size=(2, 50))
    ax = sns.kdeplot(
        x=x,
        y=y,
        # cmap=sns.cubehelix_palette(start=0.9, light=2, as_cmap=True),
        # cmap='twilight',
        cmap="PuBu",
        # color='#7EB5D6',
        fill=True,
        clip=(-5, 5),
        bw_adjust=bw_adjust,
        # cut=10,
        thresh=0,
        levels=15,
        # ax=ax,
        cbar=True,
    )
    ax.set_axis_off()

    # # Seems some wrong with scalebar, so draw it with adobe illustration
    # a = 43.2906  # 单位 A, 1 A = 0.1 nm = 100 pm
    # dx = a / 800 / 10  # 每个像素 dx pm
    # length_fraction = 10 / a  # 比例尺长 1 nm
    # scalebar = ScaleBar(dx=dx, units="um", length_fraction=length_fraction)
    # # 添加比例尺
    # ax.add_artist(scalebar)

    # plt.savefig(f"./eps/{filename}.eps", format="eps", bbox_inches="tight")
    plt.savefig(f"./{filename}.png", format="png", dpi=600, bbox_inches="tight")

    # plt.show()


# show(
#     np.sum(agc_data, axis=2, keepdims=False),
#     lowdark=False,
#     filename=f"{cifid}.zaxis",
#     emax=10,
#     figsize=(a, b),
# )
# show(
#     np.sum(agc_data, axis=0, keepdims=False),
#     lowdark=False,
#     filename=f"{cifid}.xaxis",
#     emax=25,
#     bw_adjust=0.43,
#     figsize=(b * math.sqrt(3) / 2, c),
# )


def distribution(agc_labels: np.array):
    data = {"agc_labels": agc_labels}
    data = pd.DataFrame(data)

    ax = sns.histplot(
        x="agc_labels",
        data=data,
        stat="probability",
        binwidth=1,
        color="#c08292",
    )

    plt.xlabel("Number of atoms", labelpad=10, fontdict={"size": 16, "family": "Arial"})
    plt.ylabel("Ratio", labelpad=10, fontdict={"size": 16, "family": "Arial"})
    plt.title(
        "Distribution of number of atoms",
        pad=20,
        fontdict={"size": 20, "family": "Arial"},
    )

    plt.savefig("agc.distribution.png", dpi=600, format="png", bbox_inches="tight")



from spbnet.data.buildData import buildData
import argparse
from spbnet.visualize.buildModalData import buildModalData
from pathlib import Path


class CLICommand:
    """
    build modal data for single cif
    """

    @staticmethod
    def add_arguments(parser: argparse.ArgumentParser):
        parser.add_argument(
            "--cif-path", type=str, help="cif-path to build modal data", required=True
        )
        parser.add_argument(
            "--target-dir", type=str, help="target-dir to build modal data", default="."
        )

    @staticmethod
    def run(args):
        buildModalData(Path(args.cif_path), Path(args.target_dir))



from typing import Any, Optional
from functools import partial

# from pytorch_lightning.utilities.types import STEP_OUTPUT
# from model.crossformer import CrossFormer
from spbnet.modules.module import CrossFormer
from spbnet.modules.heads import RegressionHead, Pooler
from spbnet.modules.optimize import set_scheduler
from spbnet.modules import objectives
from torch import nn
from spbnet.data.dataset import Dataset
import pytorch_lightning as pl
import pandas as pd
import yaml
import torch
from pytorch_lightning.callbacks import ModelCheckpoint
from pytorch_lightning.callbacks import LearningRateMonitor
from torch.utils.data import DataLoader
from pathlib import Path
from multiprocessing import Process
from pytorch_lightning.loggers import TensorBoardLogger
from spbnet.utils.echo import err, warn, param, title


cur_dir = Path(__file__).parent


def r2_score(y_true, y_pred):
    # 计算总平均值
    mean_true = torch.mean(y_true)

    # 计算总平方和
    total_ss = torch.sum((y_true - mean_true) ** 2)

    # 计算残差平方和
    resid_ss = torch.sum((y_true - y_pred) ** 2)

    # 计算R2分数
    r2 = 1 - (resid_ss / total_ss)

    return r2


class CrossFormerTrainer(pl.LightningModule):
    def __init__(self, config: dict):
        super(CrossFormerTrainer, self).__init__()
        self.save_hyperparameters()

        self.model_config = config
        self.optimizer_config = config

        model_config = self.model_config
        self.model = CrossFormer(model_config)
        # print(self.model)
        # self.model = AtomFormer(config)

        self.mean = config["mean"]
        self.std = config["std"]

        # pooler
        self.pooler = Pooler(model_config["hid_dim"])
        self.pooler.apply(objectives.init_weights)

        self.head = RegressionHead(model_config["hid_dim"])
        self.head.apply(objectives.init_weights)

        self.mse_loss = nn.MSELoss()
        self.mae_loss = nn.L1Loss()

        # mae
        self.min_mae = 1e5

    def training_step(self, batch, batch_idx) -> torch.Tensor:
        lrs = self.lr_schedulers().get_last_lr()
        lr_avg = sum(lrs)
        lr_avg /= len(lrs)
        self.log("lr", lr_avg)

        # [batch_size, max_token_len]
        value = batch["target"]
        value = (value - self.mean) / self.std

        feat = self.model(batch)

        # regression
        cls_feat = feat["cls_feat"]
        cls_feat = self.pooler(cls_feat)
        reg = self.head(cls_feat)  # [batch_size, 1]
        reg = reg.reshape(reg.shape[0])  # [batch_size]
        value = value.reshape(value.shape[0])  # [batch_size]
        mse = self.mse_loss(reg, value)
        mae = self.mae_loss(reg, value)
        self.log("train_mse", mse, batch_size=value.shape[0], sync_dist=True)
        self.log("train_mae", mae, batch_size=value.shape[0], sync_dist=True)

        # raw
        target = batch["target"]
        pred = reg * self.std + self.mean
        mse_raw = torch.mean((target - pred) ** 2)
        mae_raw = torch.mean(torch.abs(target - pred))
        r2 = r2_score(target, pred)
        self.log(
            "train_mse_raw",
            mse_raw,
            batch_size=value.shape[0],
            sync_dist=True,
        )
        self.log("train_mae_raw", mae_raw, batch_size=value.shape[0], sync_dist=True)
        self.log("train_r2", r2, batch_size=value.shape[0], sync_dist=True)

        # norm_mse = self.mse_loss(reg, norm_value)
        # loss = norm_mse
        loss = mse
        # loss = bert_loss
        return loss

    def on_validation_epoch_start(self) -> None:
        self.validation_items = []
        return super().on_validation_epoch_start()

    def validation_step(self, batch, batch_idx) -> torch.Tensor:
        # [batch_size, max_token_len]
        target = batch["target"]

        feat = self.model(batch)
        cls_feat = feat["cls_feat"]
        cls_feat = self.pooler(cls_feat)
        reg = self.head(cls_feat)  # [batch_size, 1]
        pred = reg * self.std + self.mean

        cifid = batch["cifid"]
        for cifid_i, target_i, pred_i in zip(cifid, target, pred):
            self.validation_items.append([cifid_i, target_i.item(), pred_i.item()])

    def on_validation_epoch_end(self) -> None:
        self.validation_df = pd.DataFrame(
            self.validation_items, columns=["cifid", "target", "predict"]
        )
        pred = torch.tensor(self.validation_df["predict"])
        target = torch.tensor(self.validation_df["target"])
        mae = torch.mean(torch.abs((pred - target)))
        mse = torch.mean((pred - target) ** 2)
        r2 = r2_score(target, pred)
        self.log("val_mae", mae)
        self.log("val_mse", mse)
        self.log("val_r2", r2)
        return super().on_validation_epoch_end()

    def on_test_epoch_start(self) -> None:
        self.test_items = []
        return super().on_test_epoch_start()

    def test_step(self, batch, batch_idx) -> torch.Tensor:
        # [batch_size, max_token_len]
        target = batch["target"]

        feat = self.model(batch)
        cls_feat = feat["cls_feat"]
        cls_feat = self.pooler(cls_feat)
        reg = self.head(cls_feat)  # [batch_size, 1]
        pred = reg * self.std + self.mean

        cifid = batch["cifid"]
        for cifid_i, target_i, pred_i in zip(cifid, target, pred):
            self.test_items.append([cifid_i, target_i.item(), pred_i.item()])

    def on_test_epoch_end(self) -> None:
        self.test_df = pd.DataFrame(
            self.test_items, columns=["cifid", "target", "predict"]
        )
        pred = torch.tensor(self.test_df["predict"])
        target = torch.tensor(self.test_df["target"])
        mae = torch.mean(torch.abs((pred - target)))
        mse = torch.mean((pred - target) ** 2)
        r2 = r2_score(target, pred)
        self.log("test_mae", mae)
        self.log("test_mse", mse)
        self.log("test_r2", r2)
        # print(f"TEST MAE: {mae.item()}, R2: {r2.item()}")
        self.test_df.to_csv(
            (Path(self.logger.log_dir) / "test_result.csv"),
            index=False,
        )
        return super().on_test_epoch_end()

    def configure_optimizers(self) -> Any:
        return set_scheduler(self)


def finetune(config_path: str):
    base_config_path = cur_dir / "config.finetune.yaml"
    with open(base_config_path, "r") as f:
        base_config: dict = yaml.load(f, Loader=yaml.FullLoader)

    with open(config_path, "r") as f:
        user_config: dict = yaml.load(f, Loader=yaml.FullLoader)

    if user_config.get("data_dir") is None:
        err(f"Please specify modal directory `data_dir`!")
        return
    if user_config.get("id_prop") is None:
        err(f"Please specify label data `id_prop`")
        return
    if user_config.get("task") is None:
        err(f"Please specify task")
        return
    if user_config.get("log_dir") is None:
        err(f"Please specify log directory")
        return

    # base_config.update(user_config)
    config = {**base_config, **user_config}

    # check
    title("FINETUNE SPBNET")
    param(**config)

    # handle
    task = config["task"]
    device = config["device"]
    id_prop_path = Path(config["id_prop"])
    ckpt_path = Path(config["ckpt"])  # TODO: CKPT
    data_dir = Path(config["data_dir"])
    log_dir = Path(config["log_dir"])

    # split train & val
    id_prop_dir = id_prop_path.parent
    train_df = pd.read_csv(
        id_prop_dir / f"{id_prop_path.stem}.train.csv", dtype={"cifid": str}
    )
    val_df = pd.read_csv(
        id_prop_dir / f"{id_prop_path.stem}.validate.csv", dtype={"cifid": str}
    )
    test_df = pd.read_csv(
        id_prop_dir / f"{id_prop_path.stem}.test.csv", dtype={"cifid": str}
    )
    train_dataset = Dataset(train_df, data_dir, config["nbr_fea_len"], task)
    val_dataset = Dataset(val_df, data_dir, config["nbr_fea_len"], task)
    test_dataset = Dataset(test_df, data_dir, config["nbr_fea_len"], task)

    # id_prop
    id_prop_df = train_df.copy()
    filter_id_prop_df = id_prop_df.dropna(subset=[task])

    # mean & std
    mean = None
    std = None
    mean = filter_id_prop_df[task].mean()
    std = filter_id_prop_df[task].std()
    config["mean"] = float(mean)
    config["std"] = float(std)

    # train
    checkpoint_callback = ModelCheckpoint(monitor="val_mae", mode="min", save_last=True)
    lr_monitor = LearningRateMonitor(logging_interval="epoch")
    logger = TensorBoardLogger(save_dir=(log_dir / task).absolute(), name="")
    trainer = pl.Trainer(
        max_epochs=config["epochs"],
        min_epochs=0,
        devices=device,
        accelerator=config["accelerator"],
        strategy=config["strategy"],
        callbacks=[checkpoint_callback, lr_monitor],
        accumulate_grad_batches=config["accumulate_grad_batches"],
        precision=config["precision"],
        log_every_n_steps=config["log_every_n_steps"],
        logger=logger,
    )
    model = CrossFormerTrainer(config)
    if not config["ckpt"] == "scratch":
        ckpt = torch.load(ckpt_path)
        model.load_state_dict(ckpt["state_dict"], strict=False)
    trainer.fit(
        model=model,
        train_dataloaders=DataLoader(
            train_dataset,
            batch_size=config["batch_size"],
            shuffle=True,
            num_workers=8,
            collate_fn=partial(Dataset.collate, img_size=config["img_size"]),
        ),
        val_dataloaders=DataLoader(
            val_dataset,
            batch_size=config["batch_size"],
            shuffle=False,
            num_workers=8,
            collate_fn=partial(Dataset.collate, img_size=config["img_size"]),
        ),
    )

    trainer.test(
        dataloaders=DataLoader(
            test_dataset,
            batch_size=config["batch_size"],
            shuffle=False,
            num_workers=8,
            collate_fn=partial(Dataset.collate, img_size=config["img_size"]),
        ),
        ckpt_path="best",
    )



from .prepare_data import make_prepared_data
from pathlib import Path
import math
import numpy as np
from ase.build import make_supercell
from ase.io import read
import shutil
import argparse
from spbnet.utils.echo import err, title, start, end


GRID = 30


def handle_griddata(file_griddata, emin=-5000.0, emax=5000, bins=101):
    griddata = np.fromfile(file_griddata.absolute(), dtype=np.float32)
    griddata = griddata.reshape(30 * 30 * 30, 20)

    griddata[griddata <= emin] = emin
    griddata[griddata > emax] = emax

    x = np.linspace(emin, emax, bins)
    griddata = np.digitize(griddata, x) + 1

    griddata = griddata.astype(np.uint8)

    return griddata


def get_atoms(cif_path: Path, tgt_dir: Path):
    try:

        def _make_supercell(atoms, cutoff):
            """
            make atoms into supercell when cell length is less than cufoff (min_length)
            """
            # when the cell lengths are smaller than radius, make supercell to be longer than the radius
            scale_abc = []
            for l in atoms.cell.cellpar()[:3]:
                if l < cutoff:
                    scale_abc.append(math.ceil(cutoff / l))
                else:
                    scale_abc.append(1)

            # make supercell
            m = np.zeros([3, 3])
            np.fill_diagonal(m, scale_abc)
            atoms = make_supercell(atoms, m)
            return atoms

        cifpath = cif_path
        cifid = cif_path.stem
        atoms = read(cifpath.absolute())
        atoms = _make_supercell(atoms, cutoff=30)
        atoms.write(tgt_dir / f"energycell/{cifid}.cif")
        atoms = read(cifpath.absolute())
        atoms = _make_supercell(atoms, cutoff=8)
        atoms.write(tgt_dir / f"supercell/{cifid}.cif")
    except Exception as e:
        raise ValueError("Build Modal Data: get atoms error", e)
    return atoms


def buildGraphAndGrid(cif_path: Path, tgt_dir: Path):
    cifpath = cif_path
    cifid = cifpath.stem
    total_dir = tgt_dir / "total"
    total_dir.mkdir(exist_ok=True)
    cifid = cifpath.stem
    make_prepared_data(cifpath, total_dir)
    # move to target folder
    for suffix in ["graphdata", "grid", "griddata"]:
        (tgt_dir / suffix).mkdir(exist_ok=True)
        if not (total_dir / f"{cifid}.{suffix}").exists():
            continue
        shutil.copy2(
            (total_dir / f"{cifid}.{suffix}"), (tgt_dir / suffix / f"{cifid}.{suffix}")
        )


def buildGriddata8(cif_path: Path, tgt_dir: Path):
    cifid = cif_path.stem
    griddata8 = handle_griddata(tgt_dir / f"griddata/{cifid}.griddata")
    np.save((tgt_dir / f"griddata8/{cifid}").absolute(), griddata8)


def buildModalData(cif_path: Path, tgt_dir: Path):
    cifid = cif_path.stem

    def ifProcessed():
        suffix_map = {
            "graphdata": "graphdata",
            "grid": "grid",
            "griddata": "griddata",
            "griddata8": "npy",
            "supercell": "cif",
            "energycell": "cif",
        }

        for suffix in [
            "graphdata",
            "grid",
            "griddata",
            "griddata8",
            "supercell",
            "energycell",
        ]:
            if not (tgt_dir / suffix / f"{cifid}.{suffix_map[suffix]}").exists():
                print(f"{suffix} not prepared")
                return False
        if not (tgt_dir / "supercell" / f"{cifid}.cif").exists():
            print(f"super cell not prepared")
            return False
        return True

    title(f"START TO BUILD MODAL DATA - {cifid}")
    start("Star to check if processed")
    if ifProcessed():
        title("HAVE PROCESSED")
        return
    end("Check processed end")
    for suffix in [
        "graphdata",
        "grid",
        "griddata",
        "griddata8",
        "supercell",
        "energycell",
        "mol",
        "attn",
    ]:
        (tgt_dir / suffix).mkdir(exist_ok=True, parents=True)
    start("Start to build graph and grid")
    buildGraphAndGrid(cif_path, tgt_dir)
    start("End to build graph and grid")
    start("Start to build griddata8")
    buildGriddata8(cif_path, tgt_dir)
    start("End to build griddata8")
    start("Start to get atoms")
    get_atoms(cif_path, tgt_dir)
    start("End to get atoms")
    title("BUILD MODAL DATA END")



from spbnet.data.calcAtomNum import calcAtomNum
import argparse


class CLICommand:
    """
    This command is used to estimate the average number of atoms in a dataset.

    This is one of the key configuration for SpbNet.
    SpbNet is based on Transformer architecture.
    The sequence need to be truncated to keep smaller than the specified max length of tokens.
    We suggest setting the "max_graph_len" hyper parameter of spbnet similar to the average atom number to maintain relatively complete information of structure.
    Low "max_graph_len" hyper parameter will result in low accuracy.

    To calculate the average atom nunber, please use:

    $ spbnet calc_atom_num PATH/TO/YOUR/CIF_DIR

    The PATH/TO/YOUR/CIF_DIR should look like:

    - root_dir
        mof1.cif
        mof2.cif
        ...
    """

    @staticmethod
    def add_arguments(parser: argparse.ArgumentParser):
        parser.add_argument(
            "--root-dir", type=str, help="root_dir containing cif files"
        )
        parser.add_argument(
            "--num", type=int, default=500, help="num of sample files, default is 500"
        )

    def run(args):
        calcAtomNum(args.root_dir, args.num)



import argparse
from pathlib import Path
from spbnet.utils.echo import title, start, end
from .buildModalData import buildModalData
from spbnet.modules.module import CrossFormer
from .utils import get_grid_data, get_graph, collate, get_atoms
import numpy as np
from pathlib import Path
from subprocess import Popen
# from rdkit import Chem
import json
from ase.io import read, write

cur_dir = Path(__file__).parent


def attn(cif_path: Path, modal_dir: Path, ckpt: Path, out_dir: Path):
    title("START TO GET ATTENTION WEIGHT")
    cifid = cif_path.stem

    percentile = 50
    # ratio = 3

    start("Start to prepare data")

    buildModalData(cif_path, modal_dir)

    end("Prepare modal data end")

    import torch
    import yaml

    start(f"Start to load weight from {ckpt}")

    ckpt = torch.load(
        ckpt,
        # f"./ckpt/p10w/{probe}.ckpt",
        map_location="cpu",
    )
    state_dict = ckpt["state_dict"]
    for key in list(state_dict.keys()):
        if not key.startswith("model."):
            del state_dict[key]
    state_dict = {key[len("model.") :]: value for key, value in state_dict.items()}
    with open((cur_dir / "config.yaml").absolute(), "r") as f:
        config = yaml.load(f, Loader=yaml.FullLoader)
    config["visualize"] = True
    model = CrossFormer(config=config)
    model.load_state_dict(state_dict, strict=True)
    model.eval()

    end("Weight load success")

    print("Start to get atttention weight")

    data_dir = modal_dir

    batch = []
    item = dict()
    item.update({"cifid": cifid, "target": 0})
    item.update(get_grid_data(data_dir, cifid))
    graph_data = get_graph(data_dir, cifid)
    item.update(graph_data)
    atom_num = min(graph_data["atom_num"].shape[0], 512)
    atom_idx = graph_data["atom_num"]
    atom_idx = atom_idx.detach().numpy()
    batch.append(item)
    batch = collate(batch)
    out = model(batch)
    sa_attn: torch.Tensor = out["sa_attn"]  # [B, lj_len, lj_len]
    ma_attn: torch.Tensor = out["ca_attn"]  # [B, lj_len, graph_len]

    potential_attn = sa_attn[0, 0].detach().numpy()
    # potential_attn = potential_attn[2:-2].detach().numpy()
    atom_attn = ma_attn[0, 0].detach().numpy()  # [graph_len]
    top_potential_indices = np.array(np.argsort(potential_attn)[-1:])
    # max_potential_attn_idx = np.argmax(potential_attn)
    # potential_atom_attn = ma_attn[0, int(max_potential_attn_idx)].detach().numpy()
    potential_atom_attn = ma_attn[0, top_potential_indices].detach().numpy()
    # potential_atom_attn = np.concatenate(
    #     [potential_atom_attn, atom_attn.reshape(1, -1)], axis=0
    # )
    potential_atom_attn = np.max(potential_atom_attn, axis=0, keepdims=False)
    # potential_atom_attn.shape, potential_atom_attn[:10], atom_attn.shape, atom_attn[
    #     :10
    # ], max_potential_attn_idx
    end(
        f"Get attention weight success, Max: {np.max(atom_attn)}, Min: {np.min(atom_attn)}, Mean: {np.mean(atom_attn)}"
    )

    start("Start to construct mol data")

    # proc = Popen(
    #     [
    #         "obabel",
    #         (modal_dir / f"supercell/{cifid}.cif").absolute(),
    #         "-O",
    #         (modal_dir / f"mol/{cifid}.mol").absolute(),
    #         "--addhydrogens",
    #     ]
    # )
    # proc.wait()
    xyz_atoms = read((modal_dir / f"supercell/{cifid}.cif").absolute())
    write((modal_dir / f"mol/{cifid}.xyz").absolute(), xyz_atoms, format="xyz")

    end("Construct mol data success")

    start("Start to get atoms")
    atoms = get_atoms((modal_dir / f"supercell/{cifid}.cif").absolute())
    atom_positions = atoms.get_positions()
    end(f"Supercell, atom num: {len(atom_positions)}")

    # # 读取MOL文件
    # mol_file_path = (modal_dir / "mol" / f"{cifid}.mol").absolute()
    # mol = Chem.MolFromMolFile(mol_file_path, sanitize=False)

    # # 获取原子坐标
    # conf = mol.GetConformer()
    # mol_positions = [list(conf.GetAtomPosition(i)) for i in range(mol.GetNumAtoms())]
    # len(mol_positions), mol_positions[:10]
    # end(f"Mol data got, atom num: {len(mol_positions)}")

    # start(f"Use template from ./template/attn.html")
    start("Start to build attention html")

    vis_attn: np.array = atom_attn
    vis_attn = vis_attn[:atom_num]
    atom_idx = atom_idx[:atom_num]
    mean, std = np.mean(vis_attn), np.std(vis_attn)
    vis_attn[np.where(atom_idx == 1)] = mean
    # ratio = 3 / (std / mean)
    # attn_mean = np.mean(vis_attn)
    # vis_attn /= attn_mean
    # vis_attn **= ratio
    # vis_attn[vis_attn > 4] = 4
    jsonData = [float(vis_attn[i]) for i in range(min(512, len(atom_positions)))]
    with open(modal_dir / f"attn/{cifid}.json", "w") as f:
        json.dump(jsonData, f)

    attn_html = None
    with open(cur_dir / "template/attn.html", "r") as f:
        attn_html = f.read()
    xyzData_str = None
    attn_str = None
    with open(modal_dir / f"mol/{cifid}.xyz", "r") as f:
        xyzData_str = f.read()
    with open(modal_dir / f"attn/{cifid}.json", "r") as f:
        attn_str = f.read()
    attn_html = attn_html.replace("__xyzData__", xyzData_str)
    attn_html = attn_html.replace("__attn__", attn_str)

    out_dir.mkdir(exist_ok=True)

    with open((out_dir / f"{cifid}.html").absolute(), "w") as f:
        f.write(attn_html)

    end("Visualization success")

    title("GOT ATTENTION SUCCESS")



# MOFTransformer version 2.0.0
import subprocess
from pathlib import Path
from .prepare_data import GRIDAY_PATH
from spbnet.utils.echo import title, err, start, end


class InstallationError(Exception):
    def __init__(self, error_message=None):
        self.error_message = error_message

    def __str__(self):
        if self.error_message:
            return f"Installation Error : {self.error_message}"
        return "Installation Error"


def install_make():
    title("INSTALL MAKE")
    start("Downloading gcc=9.5.0")
    ps = subprocess.run("conda install -c conda-forge gcc=9.5.0 -y".split())
    if ps.returncode:
        raise InstallationError(ps.stderr)
    else:
        end("Successfully download")
    start("Downloading gxx=9.5.0")
    ps = subprocess.run("conda install -c conda-forge gxx=9.5.0 -y".split())
    if ps.returncode:
        raise InstallationError(ps.stderr)
    else:
        end("Successfully download")
    start("Downloading make=4.2.1")
    ps = subprocess.run("conda install -c anaconda make=4.2.1 -y".split())
    if ps.returncode:
        raise InstallationError(ps.stderr)
    else:
        end("Successfully download")
    title("INSTALL MAKE END")


def make_griday():
    title("MAKEING GRIDAY")
    dir_griday = Path(GRIDAY_PATH).parent.parent
    if not dir_griday.exists():
        raise InstallationError(f"Invalid path specified : {dir_griday}")
    start("Makeing GRIDAY")
    ps = subprocess.run(["make"], cwd=dir_griday)
    if ps.returncode:
        raise InstallationError(ps.stderr)
    ps = subprocess.run(["make"], cwd=dir_griday / "scripts")
    if ps.returncode:
        raise InstallationError(ps.stderr)
    end("Successfully make")
    if not Path(GRIDAY_PATH).exists():
        raise InstallationError(f"GRIDAY is not installed. Please try again.")

    start("Checking GIRDAY")
    ps = subprocess.run(
        [str(GRIDAY_PATH)],
        cwd=dir_griday,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    if (
        ps.stderr
        == b"./make_egrid spacing atom_type force_field input_cssr egrid_stem\n"
    ):
        end(f"GRIDAY is installed to {dir_griday}")
    else:
        err(ps.stdout, ps.stderr)
        err(
            f"GRIDAY does not installed correctly. Please uninstall griday and re-install."
        )
    title("MAKE GRIDAY END")


def uninstall_griday():
    dir_griday = Path(GRIDAY_PATH).parent.parent
    if not dir_griday.exists():
        raise InstallationError(f"Invalid path specified : {dir_griday}")
    title("UNINSTALLING GRIDAY")
    ps = subprocess.run(["make", "clean"], cwd=dir_griday)
    if ps.returncode:
        raise InstallationError(ps.stderr)
    ps = subprocess.run(["make", "clean"], cwd=dir_griday / "scripts")
    if ps.returncode:
        raise InstallationError(ps.stderr)

    if not Path(GRIDAY_PATH).exists():
        end(f"GRIDAY is uninstalled")
    else:
        raise InstallationError()
    title("UNINSTALL GRIDAY END")



from .prepare_data import make_prepared_data
from .float2int8 import float2int8
from pathlib import Path
from multiprocessing import Process
import shutil
import argparse
from tqdm import tqdm
from spbnet.utils.echo import title, err, param, start, end


def buildData(
    root_dir, cif_dir, target_dir, n_process, crystal_max_length, max_num_unique_atoms
):
    def buildGraphAndGrid(cif_path: Path, target_dir: Path):
        cifid = cif_path.stem

        def ifProcessed():
            suffix_map = {
                "graphdata": "graphdata",
                "grid": "grid",
                "griddata": "griddata",
            }
            for suffix in ["graphdata", "grid", "griddata"]:
                if not (
                    target_dir / suffix / f"{cifid}.{suffix_map.get(suffix)}"
                ).exists():
                    return False
            return True

        if ifProcessed():
            return

        total_dir = target_dir / "total"
        total_dir.mkdir(exist_ok=True)
        make_prepared_data(
            cif_path,
            total_dir,
            max_length=crystal_max_length,
            max_num_unique_atoms=max_num_unique_atoms,
        )
        # move to target folder
        for suffix in ["graphdata", "grid", "griddata"]:
            (target_dir / suffix).mkdir(exist_ok=True)
            if not (total_dir / f"{cifid}.{suffix}").exists():
                continue
            shutil.move(
                (total_dir / f"{cifid}.{suffix}"),
                (target_dir / suffix / f"{cifid}.{suffix}"),
            )

    def process_cif(cifpath: Path, target_dir: Path):
        buildGraphAndGrid(cifpath, target_dir)

    def process_multi_cif(cifpaths, target_dir: Path):
        for cifpath in tqdm(cifpaths):
            try:
                process_cif(cifpath, target_dir)
            except Exception as e:
                err(f"Error when processing {cifpath.stem}: {e}")

    root_dir = Path(root_dir)
    cif_dir = root_dir / cif_dir
    target_dir = root_dir / target_dir
    target_dir.mkdir(exist_ok=True)

    cif_paths = list(cif_dir.iterdir())
    cif_paths = list(filter(lambda item: item.suffix == ".cif", cif_paths))

    title("DATA PREPROCESSING")
    param(
        cif_dir=cif_dir.absolute(),
        target_dir=target_dir.absolute(),
        cif_num=len(cif_paths),
    )

    start("Process [1/2]: prepare modal data")

    process_num = n_process

    cif_num_per_process = len(cif_paths) // process_num + 1
    processes = []
    for i in range(process_num):
        cif_paths_i = cif_paths[i * cif_num_per_process : (i + 1) * cif_num_per_process]
        process = Process(target=process_multi_cif, args=(cif_paths_i, target_dir))
        process.start()
        processes.append(process)
    for process in processes:
        process.join()

    end("Process [1/2] end")

    start("Process [2/2]: float2int8")

    float2int8(target_dir, n_process=process_num)

    end("Process [2/2] end")

    title("DATA PREPROCESSING END")



import torch.nn as nn



class TransformerDecoderLayer(nn.TransformerDecoderLayer):
    # self-attention block
    def _sa_block_attn(
        self, x, attn_mask, key_padding_mask
    ):
        x, attn = self.self_attn(
            x,
            x,
            x,
            attn_mask=attn_mask,
            key_padding_mask=key_padding_mask,
            need_weights=True,
        )
        return self.dropout1(x), attn

    # multihead attention block
    def _mha_block_attn(
        self,
        x,
        mem,
        attn_mask,
        key_padding_mask,
    ):
        x, attn = self.multihead_attn(
            x,
            mem,
            mem,
            attn_mask=attn_mask,
            key_padding_mask=key_padding_mask,
            need_weights=True,
        )
        return self.dropout2(x), attn



# MOFTransformer version 2.0.0
import torch.nn as nn

from transformers.models.bert.modeling_bert import (
    BertConfig,
    BertPredictionHeadTransform,
)


class Pooler(nn.Module):
    def __init__(self, hidden_size, index=0):
        super().__init__()
        self.dense = nn.Linear(hidden_size, hidden_size)
        self.activation = nn.Tanh()
        self.index = index

    def forward(self, hidden_states):
        first_token_tensor = hidden_states[:, self.index]
        pooled_output = self.dense(first_token_tensor)
        pooled_output = self.activation(pooled_output)
        return pooled_output


class GGMHead(nn.Module):
    """
    head for Graph Grid Matching
    """

    def __init__(self, hidden_size):
        super().__init__()
        self.fc = nn.Linear(hidden_size, 2)

    def forward(self, x):
        x = self.fc(x)
        return x


class MPPHead(nn.Module):
    """
    head for Masked Patch Prediction (regression version)
    """

    def __init__(self, hid_dim):
        super().__init__()

        bert_config = BertConfig(
            hidden_size=hid_dim,
        )
        self.transform = BertPredictionHeadTransform(bert_config)
        self.decoder = nn.Linear(hid_dim, 101 + 2)  # bins

    def forward(self, x):  # [B, max_len, hid_dim]
        x = self.transform(x)  # [B, max_len, hid_dim]
        x = self.decoder(x)  # [B, max_len, bins]
        return x


class MTPHead(nn.Module):
    """
    head for MOF Topology Prediction
    """

    def __init__(self, hid_dim):
        super().__init__()
        self.fc = nn.Linear(hid_dim, 1397)  # len(assets/topology.json)

    def forward(self, x):
        x = self.fc(x)
        return x


class VFPHead(nn.Module):
    """
    head for Void Fraction Prediction
    """

    def __init__(self, hid_dim):
        super().__init__()
        self.bn = nn.BatchNorm1d(hid_dim)
        self.fc = nn.Linear(hid_dim, 1)

    def forward(self, x):
        x = self.bn(x)
        x = self.fc(x)
        return x


class RegressionHead(nn.Module):
    """
    head for Regression
    """

    def __init__(self, hid_dim):
        super().__init__()
        self.fc = nn.Linear(hid_dim, 1)

    def forward(self, x):
        x = self.fc(x)
        return x


class ClassificationHead(nn.Module):
    """
    head for Classification
    """

    def __init__(self, hid_dim, n_classes):
        super().__init__()

        if n_classes == 2:
            self.fc = nn.Linear(hid_dim, 1)
            self.binary = True
        else:
            self.fc = nn.Linear(hid_dim, n_classes)
            self.binary = False

    def forward(self, x):
        x = self.fc(x)

        return x


class MOCHead(nn.Module):
    """
    head for Metal Organic Classification
    """

    def __init__(self, hid_dim):
        super().__init__()
        self.fc = nn.Linear(hid_dim, 1)

    def forward(self, x):
        """
        :param x: graph_feats [B, graph_len, hid_dim]
        :return: [B, graph_len]
        """
        x = self.fc(x)  # [B, graph_len, 1]
        x = x.squeeze(dim=-1)  # [B, graph_len]
        return x



# MOFTransformer version 2.1.0
import torch
import torch.nn as nn

from spbnet.modules import objectives, heads
from spbnet.modules.cgcnn import GraphEmbeddings
from spbnet.modules.vision_transformer_3d import VisionTransformer3D
from spbnet.modules.transformer import TransformerDecoderLayer
from torch.nn import (
    TransformerDecoder,
    # TransformerDecoderLayer,
    TransformerEncoder,
    TransformerEncoderLayer,
)

import math


class CrossFormer(nn.Module):
    def __init__(self, config):
        super().__init__()
        # self.save_hyperparameters()
        self.config = config
        self.config["potential"] = self.config["lj"] or self.config["coulomb"]

        self.max_grid_len = config["max_grid_len"]
        # self.vis = config["visualize"]

        # graph embedding
        self.graph_embeddings = GraphEmbeddings(
            atom_fea_len=config["atom_fea_len"],
            nbr_fea_len=config["nbr_fea_len"],
            max_graph_len=config["max_graph_len"],
            hid_dim=config["hid_dim"],
            vis=config["visualize"],
            # n_conv=config["nlayers"][0],
        )
        self.graph_embeddings.apply(objectives.init_weights)

        # energy embedding
        if config["lj"]:
            self.lj_vit3d = VisionTransformer3D(
                img_size=config["img_size"]["lj"],
                patch_size=config["patch_size"]["lj"],
                in_chans=config["in_chans"],
                embed_dim=config["hid_dim"],
                depth=1,
                num_heads=config["nhead"],
                # mlp_ratio=config["mlp_ratio"],
                # drop_rate=config["drop_rate"],
                # mpp_ratio=config["mpp_ratio"],
            )

        if config["coulomb"]:
            self.coulomb_vit3d = VisionTransformer3D(
                img_size=config["img_size"]["coulomb"],
                patch_size=config["patch_size"]["coulomb"],
                in_chans=config["in_chans"],
                embed_dim=config["hid_dim"],
                depth=1,
                num_heads=config["nhead"],
                # mlp_ratio=config["mlp_ratio"],
                # drop_rate=config["drop_rate"],
                # mpp_ratio=config["mpp_ratio"],
            )

        if config["structure"]:
            structure_encoder_layer = TransformerEncoderLayer(
                config["hid_dim"],
                config["nhead"],
                config["hid_dim"],
                config["dropout"],
                batch_first=True,
            )
            self.structure_encoder = TransformerEncoder(
                structure_encoder_layer, config["nlayers"]["structure"]
            )

        # potential
        if self.config["potential"] and not self.config["structure"]:
            potential_encoder_layer = TransformerEncoderLayer(
                config["hid_dim"],
                config["nhead"],
                config["hid_dim"],
                config["dropout"],
                batch_first=True,
            )
            self.potential_encoder = TransformerEncoder(
                potential_encoder_layer, config["nlayers"]["potential"]
            )
        elif self.config['potential']:
            potential_decoder_layer = TransformerDecoderLayer(
                config["hid_dim"],
                config["nhead"],
                config["hid_dim"],
                config["dropout"],
                batch_first=True,
            )
            self.potential_decoder = TransformerDecoder(
                potential_decoder_layer, config["nlayers"]["potential"]
            )

        # potential correction
        if self.config["correction"]:
            self.potential_mapper = nn.Linear(18, 1, bias=False)
            self.potential_mapper.apply(objectives.init_weights)

        # token type embeddings
        # LJ & coulomb
        self.token_type_embeddings = nn.Embedding(2, config["hid_dim"])
        self.token_type_embeddings.apply(objectives.init_weights)

        # head
        self.pooler = heads.Pooler(config["hid_dim"])
        self.pooler.apply(objectives.init_weights)

        # class token
        self.cls_embeddings = nn.Linear(1, config["hid_dim"])
        self.cls_embeddings.apply(objectives.init_weights)

        # volume token
        self.volume_embeddings = nn.Linear(1, config["hid_dim"])
        self.volume_embeddings.apply(objectives.init_weights)

        # sep token
        self.sep_embeddings = nn.Linear(1, config["hid_dim"])
        self.sep_embeddings.apply(objectives.init_weights)

        # # regression
        # self.regression_head = heads.RegressionHead(config["hid_dim"])

    def forward(
        self,
        batch,
        mask_grid=False,
    ):
        cifid = batch["cifid"]

        # atom_num_masked = batch["atom_num_masked"]  # [N']
        # atom_num_mask = batch["atom_num_mask"]  # [N']

        atom_num = batch["atom_num"]  # [N']
        nbr_idx = batch["nbr_idx"]  # [N', M]
        nbr_fea = batch["nbr_fea"]  # [N', M, nbr_fea_len]
        crystal_atom_idx = batch["crystal_atom_idx"]  # list [B]
        uni_idx = batch["uni_idx"]  # list [B]
        uni_count = batch["uni_count"]  # list [B]

        # grid = batch["grid"]  # [B, C, H, W, D, 20]
        lj = batch["lj"]  # [B, H, W, D]
        corr = batch["corr"]  # [B, H, W, D, 18]
        volume = batch["volume"]  # list [B]

        coulomb = None
        # coulomb = batch.get("coulomb")  # [B, H, W, D]
        # coulomb.unsqueeze_(1) # [B, C, H, W, D]

        # grid correction
        if self.config["correction"]:
            corr = self.potential_mapper(corr).squeeze(-1)  # [B, H, W, D]
            grid = (lj + corr).unsqueeze(1)  # [B, 1, H, W, D]
        else:
            grid = lj
            grid = grid.unsqueeze(1)
        # grid = (bias + grid_delta).unsqueeze(1)

        # graph embeds
        (
            graph_embed,  # [B, max_graph_len, hid_dim],
            graph_mask,  # [B, max_graph_len],
        ) = self.graph_embeddings(
            atom_num=atom_num,
            nbr_idx=nbr_idx,
            nbr_fea=nbr_fea,
            crystal_atom_idx=crystal_atom_idx,
            uni_idx=uni_idx,
            uni_count=uni_count,
        )

        if self.config["lj"]:
            # lj embed
            (
                lj_embed,  # [B, max_grid_len+1, hid_dim]
                lj_mask,  # [B, max_grid_len+1]
                _,  # [B, grid+1, C] if mask_image == True
            ) = self.lj_vit3d.visual_embed(
                grid,
                max_image_len=self.max_grid_len,
                mask_it=mask_grid,
            )

        if self.config["coulomb"]:
            # coulomb embed
            (
                coulomb_embed,  # [B, max_grid_len+1, hid_dim]
                coulomb_mask,  # [B, max_grid_len+1]
                _,  # [B, grid+1, C] if mask_image == True
            ) = self.coulomb_vit3d.visual_embed(
                coulomb,
                max_image_len=self.max_grid_len,
                mask_it=mask_grid,
            )

        # fusion
        batch_size = lj_embed.shape[0]
        device = lj_embed.device

        # volume embed
        volume = torch.FloatTensor(volume).to(lj_embed)  # [B]
        volume_embed = self.volume_embeddings(volume[:, None, None])  # [B, 1, hid_dim]
        volume_mask = torch.ones(volume.shape[0], 1).to(lj_mask)

        # cls_embed
        cls_token = torch.zeros(batch_size).to(device)  # [B]
        cls_embed = self.cls_embeddings(cls_token[:, None, None])  # [B, 1, hid_dim]
        cls_mask = torch.ones(batch_size, 1).to(device)  # [B, 1]

        # sep_embed
        sep_token = torch.zeros(batch_size).to(device)  # [B]
        sep_embed = self.sep_embeddings(sep_token[:, None, None])  # [B, 1, hid_dim]
        sep_mask = torch.ones(batch_size, 1).to(device)  # [B, 1]

        # add token_type_embeddings
        if self.config["lj"]:
            lj_embed = lj_embed + self.token_type_embeddings(
                torch.zeros_like(lj_mask, device=device).long()
            )
        if self.config["coulomb"]:
            coulomb_embed = coulomb_embed + self.token_type_embeddings(
                torch.ones_like(coulomb_mask, device=device).long()
            )

        # ablation
        if self.config["structure"] and not self.config["potential"]:
            graph_embed = torch.cat([cls_embed, graph_embed], dim=1)
            graph_mask = torch.cat([cls_mask, graph_mask], dim=1)
            feat = self.structure_encoder(graph_embed, src_key_padding_mask=graph_mask)
            cls_feat = feat[:, 0:1, :]
            return {
                "cls_feat": cls_feat,
            }

        if self.config["potential"] and not self.config["structure"]:
            if self.config["lj"] and self.config["coulomb"]:
                potential_embed = torch.cat(
                    [
                        cls_embed,
                        lj_embed,
                        sep_embed,
                        coulomb_embed,
                        sep_embed,
                        volume_embed,
                    ],
                    dim=1,
                )
                potential_mask = torch.cat(
                    [cls_mask, lj_mask, sep_mask, coulomb_mask, sep_mask, volume_mask],
                    dim=1,
                )  # [B, max_grid_len+2]
            elif self.config["lj"]:
                potential_embed = torch.cat(
                    [cls_embed, lj_embed, sep_embed, volume_embed],
                    dim=1,
                )
                potential_mask = torch.cat(
                    [cls_mask, lj_mask, sep_mask, volume_mask], dim=1
                )  # [B, max_grid_len+2]
            else:
                potential_embed = torch.cat(
                    [cls_embed, coulomb_embed, sep_embed, volume_embed], dim=1
                )  # [B, max_grid_len+2, hid_dim]
                potential_mask = torch.cat(
                    [cls_mask, coulomb_mask, sep_mask, volume_mask], dim=1
                )  # [B, max_grid_len+2]
            feat = self.potential_encoder(
                potential_embed, src_key_padding_mask=potential_mask
            )
            cls_feat = feat[:, 0:1, :]
            return {
                "cls_feat": cls_feat,
            }

        if self.config["lj"] and self.config["coulomb"]:
            # potential_embed = torch.cat(
            #     [
            #         cls_embed,
            #         lj_embed,
            #         sep_embed,
            #         coulomb_embed,
            #         sep_embed,
            #         volume_embed,
            #     ],
            #     dim=1,
            # )  # [B, max_grid_len+2, hid_dim]
            # potential_mask = torch.cat(
            #     [cls_mask, lj_mask, sep_mask, coulomb_mask, sep_mask, volume_mask],
            #     dim=1,
            # )  # [B, max_grid_len+2]
            if lj_embed.shape != coulomb_embed.shape:
                raise ValueError(
                    f"lj img {self.config['img_size']['lj'], self.config['patch_size']['lj']} while coulomb img {self.config['img_size']['coulomb'], self.config['patch_size']['coulomb']}"
                )
            potential_embed = lj_embed + coulomb_embed
            potential_mask = lj_mask.bool() | coulomb_mask.bool()
            potential_embed = torch.cat(
                [cls_embed, potential_embed, sep_embed, volume_embed], dim=1
            )
            potential_mask = torch.cat(
                [cls_mask, potential_mask, sep_mask, volume_mask], dim=1
            )
        elif self.config["lj"] and not self.config["coulomb"]:
            potential_embed = torch.cat(
                [cls_embed, lj_embed, sep_embed, volume_embed], dim=1
            )  # [B, max_grid_len+2, hid_dim]
            potential_mask = torch.cat(
                [cls_mask, lj_mask, sep_mask, volume_mask], dim=1
            )  # [B, max_grid_len+2]

        structure_mem, structure_mask = None, None
        structure_mem = self.structure_encoder(
            graph_embed, src_key_padding_mask=graph_mask
        )
        structure_mask = graph_mask

        sa_attn, ca_attn = None, None
        if not self.config["visualize"]:
            feat = self.potential_decoder(
                tgt=potential_embed,
                tgt_key_padding_mask=potential_mask,
                memory=structure_mem,
                memory_key_padding_mask=structure_mask,
                tgt_mask=None,
            )
        else:
            feat = potential_embed
            for idx, layer in enumerate(self.potential_decoder.layers):
                if idx == len(self.potential_decoder.layers) - 1:
                    x = feat
                    tgt_key_padding_mask = potential_mask
                    tgt_mask = None
                    x_add, sa_attn = layer._sa_block_attn(
                        x, tgt_mask, tgt_key_padding_mask
                    )
                    x = layer.norm1(x + x_add)
                    x_add, ca_attn = layer._mha_block_attn(
                        x,
                        structure_mem,
                        None,
                        structure_mask,
                    )
                    x = layer.norm2(x + x_add)
                    x = layer.norm3(x + layer._ff_block(x))
                    feat = x
                else:
                    feat = layer(
                        tgt=feat,
                        tgt_key_padding_mask=potential_mask,
                        memory=structure_mem,
                        memory_key_padding_mask=structure_mask,
                        tgt_mask=None,
                    )

        cls_feat = feat[:, 0:1, :]

        potential_feat = feat[
            :, 2:-2, :
        ]  # cls_embed, lj_cls_embed, sep_embed, volume_embed

        # return cls_feat, smiles_feat, smiles_token_mask
        return {
            "cifid": cifid,
            "cls_feat": cls_feat,
            "sa_attn": sa_attn,
            "ca_attn": ca_attn,
            "potential_feat": potential_feat,
            "structure_feat": structure_mem,
            "structure_mask": graph_mask
        }

    def mask_tokens(self, src, mask_prob, src_padding_mask):
        mask = torch.zeros_like(src).bernoulli_(mask_prob).bool()
        if src_padding_mask is not None:
            mask = mask & ~src_padding_mask
        src = src.masked_fill(mask, 5)
        # mask = mask | src_padding_mask
        return mask, src

    def generate_square_subsequent_mask(self, sz, DEVICE="cuda"):
        diag = torch.triu(torch.ones((sz, sz), device=DEVICE), diagonal=1)
        diag[0:1, :] = torch.zeros([1, sz], device=DEVICE)
        diag[0, 0] = 1
        mask = diag == 1
        return mask

    def get_token_mask(self, src):
        padding_mask = src == self.PAD_IDX
        token_mask, src = self.mask_tokens(src, 0.01, padding_mask)
        return token_mask, padding_mask, src



import argparse

parser = argparse.ArgumentParser()
parser.add_argument("--cif", type=str)
args = parser.parse_args()
cifid = args.cif

# CO2
ckpt = "spbnet.180k"
# cifid = "cooperative"  # abb3976_data_s1

print(f"CIFID: {cifid}")

from .buildModalData import process

print("Start to prepare data")

process(cifid)

print("Prepare modal data end")

# %% [markdown]
# ## 获取注意力分数

# %%
import torch
import yaml
import torch.nn as nn

# from modules.module_reg import CrossFormer
from modules.module import CrossFormer
from modules.heads import RegressionHead, ClassificationHead, Pooler
from modules import objectives
import pytorch_lightning as pl
from pathlib import Path
from spbnet.utils.echo import start, end, title

cur_dir = Path(__file__).parent


class CrossFormerTrainer(pl.LightningModule):
    def __init__(self, config: dict):
        super(CrossFormerTrainer, self).__init__()
        self.save_hyperparameters()

        self.config = config
        self.model_config = config
        self.optimizer_config = config

        model_config = self.model_config
        self.model = CrossFormer(model_config)
        # print(self.model)

        # pooler
        self.pooler = Pooler(model_config["hid_dim"])
        self.pooler.apply(objectives.init_weights)

        # void fraction prediction
        self.vfp_head = RegressionHead(model_config["hid_dim"])
        self.vfp_head.apply(objectives.init_weights)

        # topo classify
        self.tc_head = ClassificationHead(model_config["hid_dim"], config["topo_num"])
        self.tc_head.apply(objectives.init_weights)

        # atom grid classify
        self.agc_head = nn.Linear(model_config["hid_dim"], 1)
        self.agc_head.apply(objectives.init_weights)

        self.mse_loss = nn.MSELoss()
        self.mae_loss = nn.L1Loss()
        self.cross_entropy = nn.CrossEntropyLoss()

    def pred_agc(self, batch) -> torch.Tensor:
        feat = self.model(batch)

        cls_feat = feat["cls_feat"]
        cls_feat = self.pooler(cls_feat)

        # atom grid
        potential_feat = feat[
            "potential_feat"
        ]  # [B, GRID / PathSize, GRID / PathSize, GRID / PathSize, hid_dim] aka [B, 10, 10, 10, 768]
        agc_pred = self.agc_head(potential_feat)

        return agc_pred


def predAgc(cifid: str, ckpt: str):
    title("START TO PRED AGC")

    start(f"Start to load weight from {ckpt}")

    ckpt = torch.load(
        ckpt,
        map_location="cpu",
    )
    state_dict = ckpt["state_dict"]
    with open((cur_dir / "./config.yaml").absolute(), "r") as f:
        config = yaml.load(f, Loader=yaml.FullLoader)
    # config["visualize"] = True
    model = CrossFormerTrainer(config=config)
    model.load_state_dict(state_dict, strict=True)
    model.eval()
    end(f"End to load weight from {ckpt}")

    # %%
    # print(model.potential_mapper.weight)
    # with torch.no_grad():
    #     model.potential_mapper.weight[:] = torch.ones((18)) * 2
    # model.potential_mapper.weight

    # %%
    from utils import get_grid_data, get_graph, collate
    import numpy as np
    from pathlib import Path

    start("Start to get pred agc")

    data_dir = Path("./modal")

    batch = []
    item = dict()
    item.update({"cifid": cifid, "target": 0})
    item.update(get_grid_data(data_dir, cifid))
    item.update(get_graph(data_dir, cifid))
    batch.append(item)
    batch = collate(batch)
    # cifiids = batch["cif_id"]  # list[str]
    # cifid = cifiids[0]
    # print(f"CIFID: {cifid}")
    agc_pred = model.pred_agc(batch)
    agc_pred = agc_pred.detach().cpu().reshape(-1)
    print(f"Agc getted: {agc_pred.shape}")
    np.save(f"./predagc/{cifid}.npy", agc_pred)



import numpy as np
from pathlib import Path
from tqdm import tqdm
from multiprocessing import Pool
from functools import partial


def handle_griddata(file_griddata, emin=-5000.0, emax=5000, bins=101):
    griddata = np.fromfile(file_griddata.absolute(), dtype=np.float32)
    griddata = griddata.reshape(30 * 30 * 30, 20)

    griddata[griddata <= emin] = emin
    griddata[griddata > emax] = emax

    x = np.linspace(emin, emax, bins)
    griddata = np.digitize(griddata, x) + 1

    griddata = griddata.astype(np.uint8)

    return griddata


def process(data_path: Path, target_dir: Path):
    arr = handle_griddata(data_path)
    np.save((target_dir / f"{data_path.stem}").absolute(), arr)


def float2int8(dataset_dir: Path, n_process: int):
    griddata_dir = dataset_dir / "griddata"
    target_dir = dataset_dir / "griddata8"
    target_dir.mkdir(exist_ok=True)

    data_paths = list(griddata_dir.iterdir())

    with Pool(n_process) as pool:
        result = list(
            tqdm(
                pool.imap_unordered(
                    partial(process, target_dir=target_dir), data_paths
                ),
                total=len(data_paths),
            )
        )



from ase.io import read
from pathlib import Path
import argparse
from tqdm import tqdm
import numpy as np
import random
from spbnet.utils.echo import title, start, end, param


def calcAtomNum(root_dir: str, num: int = 500):
    title("CALC ATOM NUM")

    start("Start to calculate atom num")
    root_dir = Path(root_dir)
    param(root_dir=root_dir.absolute(), sample_num=num)

    cif_file_paths = list(root_dir.iterdir())
    random.shuffle(cif_file_paths)
    cif_file_paths = cif_file_paths[:num]
    atom_nums = []
    for cif_file_path in tqdm(cif_file_paths):
        try:
            atoms = read(cif_file_path.absolute())
            atom_nums.append(len(atoms.get_atomic_numbers()))
        except:
            continue
    atom_nums = np.array(atom_nums)

    mean, std = atom_nums.mean(), atom_nums.std()
    end(
        f'Calc end, mean={mean}, std={std}. Suggest to use {mean} or {mean+std} as the "max_graph_len" hyper parameter.'
    )
    title("CALC ATOM NUM END")



from spbnet.data.checkData import checkData
import argparse


class CLICommand:
    """
    This command is used to check if the modal data (build by command `spbnt build-data`) is correct.

    If the data are properly built. The directory should look like:

    - root_dir
        - cif
            xxx.cif
        - spbnet
            - graphdata
                xxx.graphdata
            - grid
                xxx.grid
            - griddata8
                xxx.npy
        benchmark.train.csv
        benchmark.validate.csv
        benchmark.test.csv

    This command will check data in this form.
    Use it like:

    $ spbnet check_data --root-dir PATH/TO/ROOT/DIR --modal-dir spbnet --prop-dir . --prop-name benchmark --split all

    --root-dir is required, others are optional. Note prop_dir is relative to root_dir.
    By default, this command will check [train, validate, test].
    You can change the behavior like "--split none" to check "benchmark.csv" or "--split other" to check "benchmark.other.csv"
    """

    @staticmethod
    def add_arguments(parser: argparse.ArgumentParser):
        parser.add_argument("--root-dir", type=str, required=True, help="")
        parser.add_argument("--modal-dir", type=str, default="spbnet")
        parser.add_argument("--prop-dir", type=str, default=".")
        parser.add_argument("--prop-name", type=str, default="benchmark")
        parser.add_argument("--split", type=str, default="all")

    @staticmethod
    def run(args):
        if args.split == "none":
            args.split = None
        checkData(
            args.root_dir, args.modal_dir, args.prop_dir, args.prop_name, args.split
        )



# MOFTransformer version 2.0.0
class CLICommand:
    """
    Install make which is needed to build GRIDAY(to generate potential basis function).
    """

    @staticmethod
    def add_arguments(parser):
        pass

    @staticmethod
    def run(args):
        from spbnet.data.install_griday import install_make

        install_make()
